# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import os.path
import tempfile
from unittest import mock

import requests

from openstack import exceptions
from openstack.image.v2 import _proxy
from openstack.image.v2 import cache as _cache
from openstack.image.v2 import image as _image
from openstack.image.v2 import image_tasks as _image_tasks
from openstack.image.v2 import member as _member
from openstack.image.v2 import metadef_namespace as _metadef_namespace
from openstack.image.v2 import metadef_object as _metadef_object
from openstack.image.v2 import metadef_resource_type as _metadef_resource_type
from openstack.image.v2 import metadef_schema as _metadef_schema
from openstack.image.v2 import schema as _schema
from openstack.image.v2 import service_info as _service_info
from openstack.image.v2 import task as _task
from openstack import proxy as proxy_base
from openstack.tests.unit.image.v2 import test_image as fake_image
from openstack.tests.unit import test_proxy_base

EXAMPLE = fake_image.EXAMPLE


class FakeResponse:
    def __init__(self, response, status_code=200, headers=None):
        self.body = response
        self.status_code = status_code
        headers = headers if headers else {'content-type': 'application/json'}
        self.headers = requests.structures.CaseInsensitiveDict(headers)

    def json(self):
        return self.body


class TestImageProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)
        self.proxy._connection = self.cloud


class TestImage(TestImageProxy):
    def test_image_import_no_required_attrs(self):
        # container_format and disk_format are required attrs of the image
        existing_image = _image.Image(id="id")
        self.assertRaises(
            exceptions.InvalidRequest,
            self.proxy.import_image,
            existing_image,
        )

    def test_image_import(self):
        original_image = _image.Image(**EXAMPLE)
        self._verify(
            "openstack.image.v2.image.Image.import_image",
            self.proxy.import_image,
            method_args=[original_image, "method"],
            method_kwargs={
                "uri": "uri",
            },
            expected_args=[self.proxy],
            expected_kwargs={
                "method": "method",
                "store": None,
                "uri": "uri",
                "remote_region": None,
                "remote_image_id": None,
                "remote_service_interface": None,
                "stores": [],
                "all_stores": None,
                "all_stores_must_succeed": None,
            },
        )

    def test_image_create_conflicting_options(self):
        exc = self.assertRaises(
            exceptions.SDKException,
            self.proxy.create_image,
            name='fake',
            filename='fake',
            data='fake',
            container='bare',
            disk_format='raw',
        )
        self.assertIn('filename and data are mutually exclusive', str(exc))

    def test_image_create(self):
        self.verify_create(
            self.proxy.create_image,
            _image.Image,
            method_kwargs={
                'name': 'fake',
                'disk_format': 'fake_dformat',
                'container_format': 'fake_cformat',
                'allow_duplicates': True,
                'is_protected': True,
            },
            expected_kwargs={
                'name': 'fake',
                'disk_format': 'fake_dformat',
                'container_format': 'fake_cformat',
                'is_protected': True,
                'owner_specified.openstack.md5': '',
                'owner_specified.openstack.object': 'images/fake',
                'owner_specified.openstack.sha256': '',
            },
        )

    def test_image_create_file_as_name(self):
        # if we pass a filename as an image name, we should upload the file
        # itself (and use the upload flow)
        with tempfile.NamedTemporaryFile() as tmpfile:
            name = os.path.basename(tmpfile.name)
            self._verify(
                'openstack.image.v2._proxy.Proxy._upload_image',
                self.proxy.create_image,
                method_kwargs={
                    'name': tmpfile.name,
                    'allow_duplicates': True,
                },
                expected_args=[
                    name,
                ],
                expected_kwargs={
                    'filename': tmpfile.name,
                    'data': None,
                    'meta': {},
                    'wait': False,
                    'timeout': 3600,
                    'validate_checksum': False,
                    'use_import': False,
                    'stores': None,
                    'import_method': None,
                    'uri': None,
                    'remote_region': None,
                    'remote_image_id': None,
                    'remote_service_interface': None,
                    'all_stores': None,
                    'all_stores_must_succeed': None,
                    'disk_format': 'qcow2',
                    'container_format': 'bare',
                    'properties': {
                        'owner_specified.openstack.md5': '',
                        'owner_specified.openstack.object': f'images/{name}',
                        'owner_specified.openstack.sha256': '',
                    },
                },
            )

        # but not if we use a directory...
        with tempfile.TemporaryDirectory() as tmpdir:
            self.verify_create(
                self.proxy.create_image,
                _image.Image,
                method_kwargs={
                    'name': tmpdir,
                    'allow_duplicates': True,
                },
                expected_kwargs={
                    'container_format': 'bare',
                    'disk_format': 'qcow2',
                    'name': tmpdir,
                    'owner_specified.openstack.md5': '',
                    'owner_specified.openstack.object': f'images/{tmpdir}',
                    'owner_specified.openstack.sha256': '',
                },
            )

    def test_image_create_checksum_match(self):
        fake_image = _image.Image(
            id="fake",
            properties={
                self.proxy._IMAGE_MD5_KEY: 'fake_md5',
                self.proxy._IMAGE_SHA256_KEY: 'fake_sha256',
            },
        )
        self.proxy.find_image = mock.Mock(return_value=fake_image)

        self.proxy._upload_image = mock.Mock()

        res = self.proxy.create_image(
            name='fake', md5='fake_md5', sha256='fake_sha256'
        )
        self.assertEqual(fake_image, res)
        self.proxy._upload_image.assert_not_called()

    def test_image_create_checksum_mismatch(self):
        fake_image = _image.Image(
            id="fake",
            properties={
                self.proxy._IMAGE_MD5_KEY: 'fake_md5',
                self.proxy._IMAGE_SHA256_KEY: 'fake_sha256',
            },
        )
        self.proxy.find_image = mock.Mock(return_value=fake_image)

        self.proxy._upload_image = mock.Mock()

        self.proxy.create_image(
            name='fake', data=b'fake', md5='fake2_md5', sha256='fake2_sha256'
        )
        self.proxy._upload_image.assert_called()

    def test_image_create_allow_duplicates_find_not_called(self):
        self.proxy.find_image = mock.Mock()

        self.proxy._upload_image = mock.Mock()

        self.proxy.create_image(
            name='fake',
            data=b'fake',
            allow_duplicates=True,
        )

        self.proxy.find_image.assert_not_called()

    def test_image_create_validate_checksum_data_binary(self):
        """Pass real data as binary"""
        self.proxy.find_image = mock.Mock()

        self.proxy._upload_image = mock.Mock()

        self.proxy.create_image(
            name='fake',
            data=b'fake',
            validate_checksum=True,
            container='bare',
            disk_format='raw',
        )

        self.proxy.find_image.assert_called_with('fake')

        self.proxy._upload_image.assert_called_with(
            'fake',
            container_format='bare',
            disk_format='raw',
            filename=None,
            data=b'fake',
            meta={},
            properties={
                self.proxy._IMAGE_MD5_KEY: '144c9defac04969c7bfad8efaa8ea194',
                self.proxy._IMAGE_SHA256_KEY: 'b5d54c39e66671c9731b9f471e585d8262cd4f54963f0c93082d8dcf334d4c78',  # noqa: E501
                self.proxy._IMAGE_OBJECT_KEY: 'bare/fake',
            },
            timeout=3600,
            validate_checksum=True,
            use_import=False,
            import_method=None,
            uri=None,
            remote_region=None,
            remote_image_id=None,
            remote_service_interface=None,
            stores=None,
            all_stores=None,
            all_stores_must_succeed=None,
            wait=False,
        )

    def test_image_create_validate_checksum_data_not_binary(self):
        self.assertRaises(
            exceptions.SDKException,
            self.proxy.create_image,
            name='fake',
            data=io.StringIO(),
            validate_checksum=True,
            container='bare',
            disk_format='raw',
        )

    def test_image_create_data_binary(self):
        """Pass binary file-like object"""
        self.proxy.find_image = mock.Mock()

        self.proxy._upload_image = mock.Mock()

        data = io.BytesIO(b'\0\0')

        self.proxy.create_image(
            name='fake',
            data=data,
            validate_checksum=False,
            container='bare',
            disk_format='raw',
        )

        self.proxy._upload_image.assert_called_with(
            'fake',
            container_format='bare',
            disk_format='raw',
            filename=None,
            data=data,
            meta={},
            properties={
                self.proxy._IMAGE_MD5_KEY: '',
                self.proxy._IMAGE_SHA256_KEY: '',
                self.proxy._IMAGE_OBJECT_KEY: 'bare/fake',
            },
            timeout=3600,
            validate_checksum=False,
            use_import=False,
            import_method=None,
            uri=None,
            remote_region=None,
            remote_image_id=None,
            remote_service_interface=None,
            stores=None,
            all_stores=None,
            all_stores_must_succeed=None,
            wait=False,
        )

    def test_image_create_protected(self):
        self.proxy.find_image = mock.Mock()

        created_image = mock.Mock(spec=_image.Image(id="id"))
        self.proxy._create = mock.Mock()
        self.proxy._create.return_value = created_image
        self.proxy._create.return_value.image_import_methods = []

        created_image.upload = mock.Mock()
        created_image.upload.return_value = FakeResponse(
            response="", status_code=200
        )

        properties = {"is_protected": True}

        self.proxy.create_image(
            name="fake",
            data="data",
            container_format="bare",
            disk_format="raw",
            **properties,
        )

        args, kwargs = self.proxy._create.call_args
        self.assertEqual(kwargs["is_protected"], True)

    def test_image_create_with_stores(self):
        self.proxy.find_image = mock.Mock()
        self.proxy._upload_image = mock.Mock()

        self.proxy.create_image(
            name='fake',
            data=b'fake',
            container='bare',
            disk_format='raw',
            use_import=True,
            stores=['cinder', 'swift'],
        )

        self.proxy.find_image.assert_called_with('fake')

        self.proxy._upload_image.assert_called_with(
            'fake',
            container_format='bare',
            disk_format='raw',
            filename=None,
            data=b'fake',
            meta={},
            properties={
                self.proxy._IMAGE_MD5_KEY: '',
                self.proxy._IMAGE_SHA256_KEY: '',  # noqa: E501
                self.proxy._IMAGE_OBJECT_KEY: 'bare/fake',
            },
            timeout=3600,
            validate_checksum=False,
            use_import=True,
            import_method=None,
            uri=None,
            remote_region=None,
            remote_image_id=None,
            remote_service_interface=None,
            stores=['cinder', 'swift'],
            all_stores=None,
            all_stores_must_succeed=None,
            wait=False,
        )

    def test_image_create_with_all_stores(self):
        self.proxy.find_image = mock.Mock()
        self.proxy._upload_image = mock.Mock()

        self.proxy.create_image(
            name='fake',
            data=b'fake',
            container='bare',
            disk_format='raw',
            use_import=True,
            all_stores=True,
            all_stores_must_succeed=True,
        )

        self.proxy.find_image.assert_called_with('fake')

        self.proxy._upload_image.assert_called_with(
            'fake',
            container_format='bare',
            disk_format='raw',
            filename=None,
            data=b'fake',
            meta={},
            properties={
                self.proxy._IMAGE_MD5_KEY: '',
                self.proxy._IMAGE_SHA256_KEY: '',  # noqa: E501
                self.proxy._IMAGE_OBJECT_KEY: 'bare/fake',
            },
            timeout=3600,
            validate_checksum=False,
            use_import=True,
            import_method=None,
            uri=None,
            remote_region=None,
            remote_image_id=None,
            remote_service_interface=None,
            stores=None,
            all_stores=True,
            all_stores_must_succeed=True,
            wait=False,
        )

    def test_image_upload_no_args(self):
        # container_format and disk_format are required args
        self.assertRaises(exceptions.InvalidRequest, self.proxy.upload_image)

    def test_image_upload(self):
        # NOTE: This doesn't use any of the base class verify methods
        # because it ends up making two separate calls to complete the
        # operation.
        created_image = mock.Mock(spec=_image.Image(id="id"))

        self.proxy._create = mock.Mock()
        self.proxy._create.return_value = created_image

        rv = self.proxy.upload_image(
            data="data", container_format="x", disk_format="y", name="z"
        )

        self.proxy._create.assert_called_with(
            _image.Image,
            container_format="x",
            disk_format="y",
            name="z",
        )
        created_image.upload.assert_called_with(self.proxy)
        self.assertEqual(rv, created_image)

    def test_image_download(self):
        original_image = _image.Image(**EXAMPLE)
        self._verify(
            'openstack.image.v2.image.Image.download',
            self.proxy.download_image,
            method_args=[original_image],
            method_kwargs={
                'output': 'some_output',
                'chunk_size': 1,
                'stream': True,
            },
            expected_args=[self.proxy],
            expected_kwargs={
                'output': 'some_output',
                'chunk_size': 1,
                'stream': True,
            },
        )

    @mock.patch("openstack.image.v2.image.Image.fetch")
    def test_image_stage(self, mock_fetch):
        image = _image.Image(id="id", status="queued")
        image.stage = mock.Mock()

        self.proxy.stage_image(image)
        mock_fetch.assert_called()
        image.stage.assert_called_with(self.proxy)

    @mock.patch("openstack.image.v2.image.Image.fetch")
    def test_image_stage_with_data(self, mock_fetch):
        image = _image.Image(id="id", status="queued")
        image.stage = mock.Mock()
        mock_fetch.return_value = image

        rv = self.proxy.stage_image(image, data="data")

        image.stage.assert_called_with(self.proxy)
        mock_fetch.assert_called()
        self.assertEqual(rv.data, "data")

    def test_image_stage_conflicting_options(self):
        image = _image.Image(id="id", status="queued")
        image.stage = mock.Mock()

        exc = self.assertRaises(
            exceptions.SDKException,
            self.proxy.stage_image,
            image,
            filename='foo',
            data='data',
        )
        self.assertIn(
            'filename and data are mutually exclusive',
            str(exc),
        )
        image.stage.assert_not_called()

    def test_image_stage_wrong_status(self):
        image = _image.Image(id="id", status="active")
        image.stage = mock.Mock()

        exc = self.assertRaises(
            exceptions.SDKException,
            self.proxy.stage_image,
            image,
            data="data",
        )
        self.assertIn(
            'Image stage is only possible for images in the queued state.',
            str(exc),
        )
        image.stage.assert_not_called()

    def test_image_delete(self):
        self.verify_delete(self.proxy.delete_image, _image.Image, False)

    def test_image_delete__ignore(self):
        self.verify_delete(self.proxy.delete_image, _image.Image, True)

    def test_delete_image__from_store(self):
        store = _service_info.Store(id='fast', is_default=True)
        store.delete_image = mock.Mock()
        image = _image.Image(id="id", status="queued")

        self.proxy.delete_image(image, store=store)

        store.delete_image.assert_called_with(
            self.proxy,
            image,
            ignore_missing=True,
        )

    @mock.patch("openstack.resource.Resource._translate_response")
    @mock.patch("openstack.proxy.Proxy._get")
    @mock.patch("openstack.image.v2.image.Image.commit")
    def test_image_update(
        self, mock_commit_image, mock_get_image, mock_transpose
    ):
        original_image = _image.Image(**EXAMPLE)
        mock_get_image.return_value = original_image
        EXAMPLE['name'] = 'fake_name'
        updated_image = _image.Image(**EXAMPLE)
        mock_commit_image.return_value = updated_image.to_dict()
        result = self.proxy.update_image(
            original_image, **updated_image.to_dict()
        )
        self.assertEqual('fake_name', result.get('name'))

    def test_image_get(self):
        self.verify_get(self.proxy.get_image, _image.Image)

    def test_images(self):
        self.verify_list(self.proxy.images, _image.Image)

    def test_add_tag(self):
        self._verify(
            "openstack.image.v2.image.Image.add_tag",
            self.proxy.add_tag,
            method_args=["image", "tag"],
            expected_args=[self.proxy, "tag"],
        )

    def test_remove_tag(self):
        self._verify(
            "openstack.image.v2.image.Image.remove_tag",
            self.proxy.remove_tag,
            method_args=["image", "tag"],
            expected_args=[self.proxy, "tag"],
        )

    def test_deactivate_image(self):
        self._verify(
            "openstack.image.v2.image.Image.deactivate",
            self.proxy.deactivate_image,
            method_args=["image"],
            expected_args=[self.proxy],
        )

    def test_reactivate_image(self):
        self._verify(
            "openstack.image.v2.image.Image.reactivate",
            self.proxy.reactivate_image,
            method_args=["image"],
            expected_args=[self.proxy],
        )

    def test_image_tasks(self):
        self.verify_list(
            self.proxy.image_tasks,
            _image_tasks.ImageTasks,
            method_kwargs={'image': 'image_1'},
            expected_kwargs={'image_id': 'image_1'},
        )


class TestMember(TestImageProxy):
    def test_member_create(self):
        self.verify_create(
            self.proxy.add_member,
            _member.Member,
            method_kwargs={"image": "test_id"},
            expected_kwargs={"image_id": "test_id"},
        )

    def test_member_delete(self):
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.remove_member,
            method_args=["member_id"],
            method_kwargs={"image": "image_id", "ignore_missing": False},
            expected_args=[_member.Member, None],
            expected_kwargs={
                "member_id": "member_id",
                "image_id": "image_id",
                "ignore_missing": False,
            },
        )

    def test_member_delete_ignore(self):
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.remove_member,
            method_args=["member_id"],
            method_kwargs={"image": "image_id"},
            expected_args=[_member.Member, None],
            expected_kwargs={
                "member_id": "member_id",
                "image_id": "image_id",
                "ignore_missing": True,
            },
        )

    def test_member_update(self):
        self._verify(
            "openstack.proxy.Proxy._update",
            self.proxy.update_member,
            method_args=['member_id', 'image_id'],
            expected_args=[_member.Member, None],
            expected_kwargs={'member_id': 'member_id', 'image_id': 'image_id'},
        )

    def test_member_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_member,
            method_args=['member_id'],
            method_kwargs={"image": "image_id"},
            expected_args=[_member.Member],
            expected_kwargs={'member_id': 'member_id', 'image_id': 'image_id'},
        )

    def test_member_find(self):
        self._verify(
            "openstack.proxy.Proxy._find",
            self.proxy.find_member,
            method_args=['member_id'],
            method_kwargs={"image": "image_id"},
            expected_args=[_member.Member, "member_id"],
            expected_kwargs={'ignore_missing': True, 'image_id': 'image_id'},
        )

    def test_members(self):
        self.verify_list(
            self.proxy.members,
            _member.Member,
            method_kwargs={'image': 'image_1'},
            expected_kwargs={'image_id': 'image_1'},
        )


class TestMetadefNamespace(TestImageProxy):
    def test_metadef_namespace_create(self):
        self.verify_create(
            self.proxy.create_metadef_namespace,
            _metadef_namespace.MetadefNamespace,
        )

    def test_metadef_namespace_delete(self):
        self.verify_delete(
            self.proxy.delete_metadef_namespace,
            _metadef_namespace.MetadefNamespace,
            False,
        )

    def test_metadef_namespace_delete__ignore(self):
        self.verify_delete(
            self.proxy.delete_metadef_namespace,
            _metadef_namespace.MetadefNamespace,
            True,
        )

    def test_metadef_namespace_get(self):
        self.verify_get(
            self.proxy.get_metadef_namespace,
            _metadef_namespace.MetadefNamespace,
        )

    def test_metadef_namespaces(self):
        self.verify_list(
            self.proxy.metadef_namespaces,
            _metadef_namespace.MetadefNamespace,
        )

    def test_metadef_namespace_update(self):
        # we're (intentionally) adding an additional field, 'namespace', to the
        # request body
        self.verify_update(
            self.proxy.update_metadef_namespace,
            _metadef_namespace.MetadefNamespace,
            method_kwargs={'is_protected': True},
            expected_kwargs={'namespace': 'resource_id', 'is_protected': True},
        )


class TestMetadefObject(TestImageProxy):
    def test_create_metadef_object(self):
        self.verify_create(
            self.proxy.create_metadef_object,
            _metadef_object.MetadefObject,
            method_kwargs={"namespace": "test_namespace_name"},
            expected_kwargs={"namespace_name": "test_namespace_name"},
        )

    def test_get_metadef_object(self):
        self.verify_get(
            self.proxy.get_metadef_object,
            _metadef_object.MetadefObject,
            method_kwargs={"namespace": "test_namespace_name"},
            expected_kwargs={
                "namespace_name": "test_namespace_name",
                'name': 'resource_id',
            },
            expected_args=[],
        )

    def test_metadef_objects(self):
        self.verify_list(
            self.proxy.metadef_objects,
            _metadef_object.MetadefObject,
            method_kwargs={"namespace": "test_namespace_name"},
            expected_kwargs={"namespace_name": "test_namespace_name"},
        )

    def test_update_metadef_object(self):
        self._verify(
            "openstack.proxy.Proxy._update",
            self.proxy.update_metadef_object,
            method_args=["test_metadef_object", "test_namespace_name"],
            method_kwargs={"name": "new_object"},
            expected_args=[
                _metadef_object.MetadefObject,
                'test_metadef_object',
            ],
            expected_kwargs={
                "name": "new_object",
                "namespace_name": "test_namespace_name",
            },
        )

    def test_delete_metadef_object(self):
        self.verify_delete(
            self.proxy.delete_metadef_object,
            _metadef_object.MetadefObject,
            False,
            method_kwargs={"namespace": "test_namespace_name"},
            expected_kwargs={"namespace_name": "test_namespace_name"},
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_delete_all_metadef_objects(self, mock_get_resource):
        fake_object = _metadef_namespace.MetadefNamespace()
        mock_get_resource.return_value = fake_object
        self._verify(
            "openstack.image.v2.metadef_namespace.MetadefNamespace.delete_all_objects",
            self.proxy.delete_all_metadef_objects,
            method_args=['test_namespace'],
            expected_args=[self.proxy],
        )
        mock_get_resource.assert_called_once_with(
            _metadef_namespace.MetadefNamespace, 'test_namespace'
        )


class TestMetadefResourceType(TestImageProxy):
    def test_metadef_resource_types(self):
        self.verify_list(
            self.proxy.metadef_resource_types,
            _metadef_resource_type.MetadefResourceType,
        )


class TestMetadefResourceTypeAssociation(TestImageProxy):
    def test_create_metadef_resource_type_association(self):
        self.verify_create(
            self.proxy.create_metadef_resource_type_association,
            _metadef_resource_type.MetadefResourceTypeAssociation,
            method_kwargs={'metadef_namespace': 'namespace_name'},
            expected_kwargs={'namespace_name': 'namespace_name'},
        )

    def test_delete_metadef_resource_type_association(self):
        self.verify_delete(
            self.proxy.delete_metadef_resource_type_association,
            _metadef_resource_type.MetadefResourceTypeAssociation,
            False,
            method_kwargs={'metadef_namespace': 'namespace_name'},
            expected_kwargs={'namespace_name': 'namespace_name'},
        )

    def test_delete_metadef_resource_type_association_ignore(self):
        self.verify_delete(
            self.proxy.delete_metadef_resource_type_association,
            _metadef_resource_type.MetadefResourceTypeAssociation,
            True,
            method_kwargs={'metadef_namespace': 'namespace_name'},
            expected_kwargs={'namespace_name': 'namespace_name'},
        )

    def test_metadef_resource_type_associations(self):
        self.verify_list(
            self.proxy.metadef_resource_type_associations,
            _metadef_resource_type.MetadefResourceTypeAssociation,
            method_kwargs={'metadef_namespace': 'namespace_name'},
            expected_kwargs={'namespace_name': 'namespace_name'},
        )


class TestSchema(TestImageProxy):
    def test_images_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_images_schema,
            expected_args=[_schema.Schema],
            expected_kwargs={
                'base_path': '/schemas/images',
                'requires_id': False,
            },
        )

    def test_image_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_image_schema,
            expected_args=[_schema.Schema],
            expected_kwargs={
                'base_path': '/schemas/image',
                'requires_id': False,
            },
        )

    def test_members_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_members_schema,
            expected_args=[_schema.Schema],
            expected_kwargs={
                'base_path': '/schemas/members',
                'requires_id': False,
            },
        )

    def test_member_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_member_schema,
            expected_args=[_schema.Schema],
            expected_kwargs={
                'base_path': '/schemas/member',
                'requires_id': False,
            },
        )


class TestTask(TestImageProxy):
    def test_task_get(self):
        self.verify_get(self.proxy.get_task, _task.Task)

    def test_tasks(self):
        self.verify_list(self.proxy.tasks, _task.Task)

    def test_task_create(self):
        self.verify_create(self.proxy.create_task, _task.Task)

    def test_wait_for_task_immediate_status(self):
        status = 'success'
        res = _task.Task(id='1234', status=status)

        result = self.proxy.wait_for_task(res, status, "failure", 0.01, 0.1)

        self.assertEqual(res, result)

    def test_wait_for_task_immediate_status_case(self):
        status = "SUCcess"
        res = _task.Task(id='1234', status=status)

        result = self.proxy.wait_for_task(res, status, "failure", 0.01, 0.1)

        self.assertEqual(res, result)

    def test_wait_for_task_error_396(self):
        # Ensure we create a new task when we get 396 error
        res = _task.Task(
            id='id',
            status='waiting',
            type='some_type',
            input='some_input',
            result='some_result',
        )

        mock_fetch = mock.Mock()
        mock_fetch.side_effect = [
            _task.Task(
                id='id',
                status='failure',
                type='some_type',
                input='some_input',
                result='some_result',
                message=_proxy._IMAGE_ERROR_396,
            ),
            _task.Task(id='fake', status='waiting'),
            _task.Task(id='fake', status='success'),
        ]

        self.proxy._create = mock.Mock()
        self.proxy._create.side_effect = [
            _task.Task(id='fake', status='success')
        ]

        with mock.patch.object(_task.Task, 'fetch', mock_fetch):
            result = self.proxy.wait_for_task(res, interval=0.01, wait=0.5)

            self.assertEqual('success', result.status)

            self.proxy._create.assert_called_with(
                mock.ANY, input=res.input, type=res.type
            )

    def test_wait_for_task_wait(self):
        res = _task.Task(id='id', status='waiting')

        mock_fetch = mock.Mock()
        mock_fetch.side_effect = [
            _task.Task(id='id', status='waiting'),
            _task.Task(id='id', status='waiting'),
            _task.Task(id='id', status='success'),
        ]

        with mock.patch.object(_task.Task, 'fetch', mock_fetch):
            result = self.proxy.wait_for_task(res, interval=0.01, wait=0.5)

            self.assertEqual('success', result.status)

    def test_tasks_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_tasks_schema,
            expected_args=[_schema.Schema],
            expected_kwargs={
                'base_path': '/schemas/tasks',
                'requires_id': False,
            },
        )

    def test_task_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_task_schema,
            expected_args=[_schema.Schema],
            expected_kwargs={
                'base_path': '/schemas/task',
                'requires_id': False,
            },
        )


class TestMisc(TestImageProxy):
    def test_stores(self):
        self.verify_list(self.proxy.stores, _service_info.Store)

    def test_import_info(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_import_info,
            method_args=[],
            method_kwargs={},
            expected_args=[_service_info.Import],
            expected_kwargs={'requires_id': False},
        )


class TestMetadefSchema(TestImageProxy):
    def test_metadef_namespace_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_namespace_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/namespace',
                'requires_id': False,
            },
        )

    def test_metadef_namespaces_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_namespaces_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/namespaces',
                'requires_id': False,
            },
        )

    def test_metadef_resource_type_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_resource_type_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/resource_type',
                'requires_id': False,
            },
        )

    def test_metadef_resource_types_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_resource_types_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/resource_types',
                'requires_id': False,
            },
        )

    def test_metadef_object_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_object_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/object',
                'requires_id': False,
            },
        )

    def test_metadef_objects_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_objects_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/objects',
                'requires_id': False,
            },
        )

    def test_metadef_property_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_property_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/property',
                'requires_id': False,
            },
        )

    def test_metadef_properties_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_properties_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/properties',
                'requires_id': False,
            },
        )

    def test_metadef_tag_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_tag_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/tag',
                'requires_id': False,
            },
        )

    def test_metadef_tags_schema_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_metadef_tags_schema,
            expected_args=[_metadef_schema.MetadefSchema],
            expected_kwargs={
                'base_path': '/schemas/metadefs/tags',
                'requires_id': False,
            },
        )


class TestCache(TestImageProxy):
    def test_image_cache_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_image_cache,
            expected_args=[_cache.Cache],
            expected_kwargs={'requires_id': False},
        )

    def test_cache_image_delete(self):
        self.verify_delete(
            self.proxy.cache_delete_image,
            _cache.Cache,
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_image_queue(self, mock_get_resource):
        fake_cache = _cache.Cache()
        mock_get_resource.return_value = fake_cache
        self._verify(
            "openstack.image.v2.cache.Cache.queue",
            self.proxy.queue_image,
            method_args=['image-id'],
            expected_args=[self.proxy, 'image-id'],
        )
        mock_get_resource.assert_called_once_with(_cache.Cache, None)

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_image_clear_cache(self, mock_get_resource):
        fake_cache = _cache.Cache()
        mock_get_resource.return_value = fake_cache
        self._verify(
            "openstack.image.v2.cache.Cache.clear",
            self.proxy.clear_cache,
            method_args=['both'],
            expected_args=[self.proxy, 'both'],
        )
        mock_get_resource.assert_called_once_with(_cache.Cache, None)

        mock_get_resource.reset_mock()
        self._verify(
            "openstack.image.v2.cache.Cache.clear",
            self.proxy.clear_cache,
            method_args=[],
            expected_args=[self.proxy, 'both'],
        )
        mock_get_resource.assert_called_once_with(_cache.Cache, None)



# Copyright (c) 2018 China Telecom Corporation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class FirewallRule(resource.Resource):
    resource_key = 'firewall_rule'
    resources_key = 'firewall_rules'
    base_path = '/fwaas/firewall_rules'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'action',
        'description',
        'destination_ip_address',
        'name',
        'destination_port',
        'enabled',
        'ip_version',
        'project_id',
        'protocol',
        'shared',
        'source_ip_address',
        'source_port',
        'firewall_policy_id',
    )

    # Properties
    #: The action that the API performs on traffic that matches the firewall
    #: rule. Valid values are allow or deny. Default is deny.
    action = resource.Body('action')
    #: The description of the firewall rule
    description = resource.Body('description')
    #: The destination IPv4 or IPv6 address or CIDR for the firewall rule.
    destination_ip_address = resource.Body('destination_ip_address')
    #: The destination port or port range for the firewall rule.
    destination_port = resource.Body('destination_port')
    #: Facilitates selectively turning off rules without having to disassociate
    #: the rule from the firewall policy
    enabled = resource.Body('enabled')
    #: The IP protocol version for the firewall rule. Valid values are 4 or 6.
    ip_version = resource.Body('ip_version')
    #: The name of the firewall rule.
    name = resource.Body('name')
    #: The ID of the project that owns the resource.
    project_id = resource.Body('project_id')
    #: The IP protocol for the firewall rule.
    protocol = resource.Body('protocol')
    #: Indicates whether this firewall rule is shared across all projects.
    shared = resource.Body('shared')
    #: The source IPv4 or IPv6 address or CIDR for the firewall rule.
    source_ip_address = resource.Body('source_ip_address')
    #: The source port or port range for the firewall rule.
    source_port = resource.Body('source_port')
    #: Summary field of a FirewallRule, composed of the protocol,
    #: source_ip_address:source_port,
    #: destination_ip_address:destination_port and action.
    summary = resource.Computed('summary', default='')
    #: The ID of the firewall policy.
    firewall_policy_id = resource.Body('firewall_policy_id')
    #: The ID of the firewall rule.
    id = resource.Body('id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Resource(resource.Resource):
    name_attribute = 'resource_name'
    resource_key = 'resource'
    resources_key = 'resources'
    base_path = '/stacks/%(stack_name)s/%(stack_id)s/resources'

    # capabilities
    allow_create = False
    allow_list = True
    allow_retrieve = False
    allow_delete = False
    allow_commit = False

    # Properties
    #: A list of dictionaries containing links relevant to the resource.
    links = resource.Body('links')
    #: ID of the logical resource, usually the literal name of the resource
    #: as it appears in the stack template.
    logical_resource_id = resource.Body(
        'logical_resource_id', alternate_id=True
    )
    #: Name of the resource.
    name = resource.Body('resource_name')
    #: ID of the physical resource (if any) that backs up the resource. For
    #: example, it contains a nova server ID if the resource is a nova
    #: server.
    physical_resource_id = resource.Body('physical_resource_id')
    #: A list of resource names that depend on this resource. This
    #: property facilitates the deduction of resource dependencies.
    #: *Type: list*
    required_by = resource.Body('required_by', type=list)
    #: A string representation of the resource type.
    resource_type = resource.Body('resource_type')
    #: A string representing the status the resource is currently in.
    status = resource.Body('resource_status')
    #: A string that explains why the resource is in its current status.
    status_reason = resource.Body('resource_status_reason')
    #: Timestamp of the last update made to the resource.
    updated_at = resource.Body('updated_time')



# Copyright (c) 2017 Red Hat, Inc.
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests import fakes
from openstack.tests.unit import base


_fake_zone_list = {
    "availabilityZoneInfo": [
        {"hosts": None, "zoneName": "az1", "zoneState": {"available": True}},
        {"hosts": None, "zoneName": "nova", "zoneState": {"available": False}},
    ]
}


class TestAvailabilityZoneNames(base.TestCase):
    def test_list_availability_zone_names(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/os-availability-zone',
                    json=_fake_zone_list,
                ),
            ]
        )

        self.assertEqual(['az1'], self.cloud.list_availability_zone_names())

        self.assert_calls()

    def test_unauthorized_availability_zone_names(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/os-availability-zone',
                    status_code=403,
                ),
            ]
        )

        self.assertEqual([], self.cloud.list_availability_zone_names())

        self.assert_calls()

    def test_list_all_availability_zone_names(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/os-availability-zone',
                    json=_fake_zone_list,
                ),
            ]
        )

        self.assertEqual(
            ['az1', 'nova'],
            self.cloud.list_availability_zone_names(unavailable=True),
        )

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack import resource
from openstack.shared_file_system.v2 import share_instance as _share_instance
from openstack.tests.functional.shared_file_system import base


class ShareInstanceTest(base.BaseSharedFileSystemTest):
    min_microversion = '2.7'

    def setUp(self):
        super().setUp()

        self.SHARE_NAME = self.getUniqueString()
        my_share = self.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.SHARE_ID = my_share.id
        instances_list = self.operator_cloud.share.share_instances()
        self.SHARE_INSTANCE_ID = None
        for i in instances_list:
            if i.share_id == self.SHARE_ID:
                self.SHARE_INSTANCE_ID = i.id

    def test_get(self):
        sot = self.operator_cloud.share.get_share_instance(
            self.SHARE_INSTANCE_ID
        )
        assert isinstance(sot, _share_instance.ShareInstance)
        self.assertEqual(self.SHARE_INSTANCE_ID, sot.id)

    def test_list_share_instances(self):
        share_instances = self.operator_cloud.share.share_instances()
        self.assertGreater(len(list(share_instances)), 0)
        for share_instance in share_instances:
            for attribute in (
                'id',
                'name',
                'created_at',
                'access_rules_status',
                'availability_zone',
            ):
                self.assertTrue(hasattr(share_instance, attribute))

    def test_reset(self):
        res = self.operator_cloud.share.reset_share_instance_status(
            self.SHARE_INSTANCE_ID, 'error'
        )
        self.assertIsNone(res)
        sot = self.operator_cloud.share.get_share_instance(
            self.SHARE_INSTANCE_ID
        )
        self.assertEqual('error', sot.status)

    def test_delete(self):
        sot = self.operator_cloud.share.get_share_instance(
            self.SHARE_INSTANCE_ID
        )
        fdel = self.operator_cloud.share.delete_share_instance(
            self.SHARE_INSTANCE_ID
        )
        resource.wait_for_delete(
            self.operator_cloud.share,
            sot,
            wait=self._wait_for_timeout,
            interval=2,
        )
        self.assertIsNone(fdel)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# API deprecation warnings
#
# These are for service-related deprecations, such as the removal of an API or
# API field due to a microversion.


class OpenStackDeprecationWarning(DeprecationWarning):
    """Base class for warnings about deprecated features in openstacksdk."""


class RemovedResourceWarning(OpenStackDeprecationWarning):
    """Indicates that a resource has been removed in newer API versions and
    should not be used.
    """


class RemovedFieldWarning(OpenStackDeprecationWarning):
    """Indicates that a field has been removed in newer API versions and should
    not be used.
    """


class LegacyAPIWarning(OpenStackDeprecationWarning):
    """Indicates an API that is in 'legacy' status, a long term deprecation."""


# Package deprecation warnings
#
# These are for SDK-specific deprecations, such as removed functions or
# function parameters.


class _RemovedInSDKWarning(PendingDeprecationWarning):
    """Indicates an argument that is deprecated for removal.

    This is a base class and should not be used directly.
    """


class RemovedInSDK50Warning(_RemovedInSDKWarning):
    """Indicates an argument that is deprecated for removal in SDK 5.0."""


class RemovedInSDK60Warning(_RemovedInSDKWarning):
    """Indicates an argument that is deprecated for removal in SDK 6.0."""


# General warnings
#
# These are usually related to misconfigurations.


class OpenStackWarning(Warning):
    """Base class for general warnings in openstacksdk."""


class ConfigurationWarning(OpenStackWarning):
    """Indicates an issue with configuration."""


class UnsupportedServiceVersion(OpenStackWarning):
    """Indicates a major version that SDK doesn't understand."""



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.network.v2 import quota as _quota
from openstack.tests.unit import base

fake_quota_set = {
    "cores": 20,
    "fixed_ips": -1,
    "floating_ips": 10,
    "injected_file_content_bytes": 10240,
    "injected_file_path_bytes": 255,
    "injected_files": 5,
    "instances": 10,
    "key_pairs": 100,
    "metadata_items": 128,
    "ram": 51200,
    "security_group_rules": 20,
    "security_groups": 45,
    "server_groups": 10,
    "server_group_members": 10,
}


class TestQuotas(base.TestCase):
    def setUp(self, cloud_config_fixture='clouds.yaml'):
        super().setUp(cloud_config_fixture=cloud_config_fixture)

    def test_update_quotas(self):
        project = self._get_project_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-quota-sets', project.project_id],
                    ),
                    json={'quota_set': fake_quota_set},
                    validate=dict(
                        json={'quota_set': {'cores': 1, 'force': True}}
                    ),
                ),
            ]
        )

        self.cloud.set_compute_quotas(project.project_id, cores=1)

        self.assert_calls()

    def test_update_quotas_bad_request(self):
        project = self._get_project_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-quota-sets', project.project_id],
                    ),
                    status_code=400,
                ),
            ]
        )

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.set_compute_quotas,
            project.project_id,
        )

        self.assert_calls()

    def test_get_quotas(self):
        project = self._get_project_data()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-quota-sets', project.project_id],
                    ),
                    json={'quota_set': fake_quota_set},
                ),
            ]
        )

        self.cloud.get_compute_quotas(project.project_id)

        self.assert_calls()

    def test_delete_quotas(self):
        project = self._get_project_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-quota-sets', project.project_id],
                    ),
                ),
            ]
        )

        self.cloud.delete_compute_quotas(project.project_id)

        self.assert_calls()

    def test_cinder_update_quotas(self):
        project = self._get_project_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_cinder_discovery_mock_dict(),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'volumev3',
                        'public',
                        append=['os-quota-sets', project.project_id],
                    ),
                    json=dict(quota_set={'volumes': 1}),
                    validate=dict(json={'quota_set': {'volumes': 1}}),
                ),
            ]
        )
        self.cloud.set_volume_quotas(project.project_id, volumes=1)
        self.assert_calls()

    def test_cinder_get_quotas(self):
        project = self._get_project_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_cinder_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'volumev3',
                        'public',
                        append=['os-quota-sets', project.project_id],
                        qs_elements=['usage=False'],
                    ),
                    json=dict(quota_set={'snapshots': 10, 'volumes': 20}),
                ),
            ]
        )
        self.cloud.get_volume_quotas(project.project_id)
        self.assert_calls()

    def test_cinder_delete_quotas(self):
        project = self._get_project_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'identity',
                        'public',
                        append=['v3', 'projects', project.project_id],
                    ),
                    json={'project': project.json_response['project']},
                ),
                self.get_cinder_discovery_mock_dict(),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'volumev3',
                        'public',
                        append=['os-quota-sets', project.project_id],
                    ),
                ),
            ]
        )
        self.cloud.delete_volume_quotas(project.project_id)
        self.assert_calls()

    def test_neutron_update_quotas(self):
        project = self.mock_for_keystone_projects(
            project_count=1, id_get=True
        )[0]
        self.register_uris(
            [
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'quotas', project.project_id],
                    ),
                    json={},
                    validate=dict(json={'quota': {'network': 1}}),
                )
            ]
        )
        self.cloud.set_network_quotas(project.project_id, network=1)
        self.assert_calls()

    def test_neutron_get_quotas(self):
        quota = {
            'subnet': 100,
            'network': 100,
            'floatingip': 50,
            'subnetpool': -1,
            'security_group_rule': 100,
            'security_group': 10,
            'router': 10,
            'rbac_policy': 10,
            'port': 500,
        }
        project = self.mock_for_keystone_projects(
            project_count=1, id_get=True
        )[0]
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'quotas', project.project_id],
                    ),
                    json={'quota': quota},
                )
            ]
        )
        received_quota = self.cloud.get_network_quotas(
            project.project_id
        ).to_dict(computed=False)
        expected_quota = _quota.Quota(**quota).to_dict(computed=False)
        received_quota.pop('id')
        received_quota.pop('name')
        expected_quota.pop('id')
        expected_quota.pop('name')

        self.assertDictEqual(expected_quota, received_quota)

        self.assert_calls()

    def test_neutron_get_quotas_details(self):
        quota_details = {
            'subnet': {'limit': 100, 'used': 7, 'reserved': 0},
            'network': {'limit': 100, 'used': 6, 'reserved': 0},
            'floatingip': {'limit': 50, 'used': 0, 'reserved': 0},
            'subnetpool': {'limit': -1, 'used': 2, 'reserved': 0},
            'security_group_rule': {'limit': 100, 'used': 4, 'reserved': 0},
            'security_group': {'limit': 10, 'used': 1, 'reserved': 0},
            'router': {'limit': 10, 'used': 2, 'reserved': 0},
            'rbac_policy': {'limit': 10, 'used': 2, 'reserved': 0},
            'port': {'limit': 500, 'used': 7, 'reserved': 0},
        }
        project = self.mock_for_keystone_projects(
            project_count=1, id_get=True
        )[0]
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=[
                            'v2.0',
                            'quotas',
                            project.project_id,
                            'details',
                        ],
                    ),
                    json={'quota': quota_details},
                )
            ]
        )
        received_quota_details = self.cloud.get_network_quotas(
            project.project_id, details=True
        )
        self.assertDictEqual(
            _quota.QuotaDetails(**quota_details).to_dict(computed=False),
            received_quota_details.to_dict(computed=False),
        )
        self.assert_calls()

    def test_neutron_delete_quotas(self):
        project = self.mock_for_keystone_projects(
            project_count=1, id_get=True
        )[0]
        self.register_uris(
            [
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'quotas', project.project_id],
                    ),
                    json={},
                )
            ]
        )
        self.cloud.delete_network_quotas(project.project_id)
        self.assert_calls()



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The openstack SDK.

:py:mod:`openstacksdk` is a client library for building applications to work
with OpenStack clouds. The project aims to provide a consistent and complete
set of interactions with OpenStack's many services, along with complete
documentation, examples, and tools.

There are three ways to interact with :py:mod:`openstacksdk`. The *clouds
layer*, the *proxy layer*, and the *resource layer*. Most users will make use
of either the *cloud layer* or *proxy layer*.

Listing flavours using the *cloud layer*::

    >>> import openstack
    >>> conn = openstack.connect(cloud='mordred')
    >>> for server in conn.list_servers():
    ...     print(server.to_dict())

Listing servers using the *proxy layer*::

    >>> import openstack
    >>> conn = openstack.connect(cloud='mordred')
    >>> for server in conn.compute.servers():
    ...     print(server.to_dict())

Listing servers using the *resource layer*::

    >>> import openstack
    >>> import openstack.compute.v2.server
    >>> conn = openstack.connect(cloud='mordred')
    >>> for server in openstack.compute.v2.server.Server.list(
    ...     session=conn.compute,
    ... ):
    ...     print(server.to_dict())

For more information, refer to the documentation found in each submodule.
"""

import argparse
import typing as ty

from openstack._log import enable_logging
import openstack.config
import openstack.connection

__all__ = [
    'connect',
    'enable_logging',
]


def connect(
    cloud: str | None = None,
    app_name: str | None = None,
    app_version: str | None = None,
    options: argparse.ArgumentParser | None = None,
    load_yaml_config: bool = True,
    load_envvars: bool = True,
    **kwargs: ty.Any,
) -> openstack.connection.Connection:
    """Create a :class:`~openstack.connection.Connection`

    :param string cloud:
        The name of the configuration to load from clouds.yaml. Defaults
        to 'envvars' which will load configuration settings from environment
        variables that start with ``OS_``.
    :param argparse.ArgumentParser options:
        An argparse ArgumentParser object. SDK-specific options will be
        registered, parsed out and used to configure the connection.
    :param bool load_yaml_config:
        Whether or not to load config settings from clouds.yaml files.
        Defaults to True.
    :param bool load_envvars:
        Whether or not to load config settings from environment variables.
        Defaults to True.
    :param kwargs:
        Additional configuration options.

    :returns: openstack.connnection.Connection
    :raises: keystoneauth1.exceptions.MissingRequiredOptions
        on missing required auth parameters
    """
    cloud_region = openstack.config.get_cloud_region(
        cloud=cloud,
        app_name=app_name,
        app_version=app_version,
        load_yaml_config=load_yaml_config,
        load_envvars=load_envvars,
        options=options,
        **kwargs,
    )
    return openstack.connection.Connection(
        config=cloud_region,
        vendor_hook=kwargs.get('vendor_hook'),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.shared_file_system.v2 import quota_class_set
from openstack.tests.unit import base

EXAMPLE = {
    "share_groups": 50,
    "gigabytes": 1000,
    "share_group_snapshots": 50,
    "snapshots": 50,
    "snapshot_gigabytes": 1000,
    "shares": 50,
    "id": "default",
    "share_networks": 10,
    "share_replicas": 100,
    "replica_gigabytes": 1000,
    "per_share_gigabytes": -1,
    "backups": 50,
    "backup_gigabytes": 1000,
}


class TestQuotaClassSet(base.TestCase):
    def test_basic(self):
        _quota_class_set = quota_class_set.QuotaClassSet()

        self.assertEqual('/quota-class-sets', _quota_class_set.base_path)
        self.assertTrue(_quota_class_set.allow_fetch)
        self.assertTrue(_quota_class_set.allow_commit)
        self.assertFalse(_quota_class_set.allow_create)
        self.assertFalse(_quota_class_set.allow_delete)
        self.assertFalse(_quota_class_set.allow_list)
        self.assertFalse(_quota_class_set.allow_head)

    def test_get_quota_class_set(self):
        _quota_class_set = quota_class_set.QuotaClassSet(**EXAMPLE)
        self.assertEqual(
            EXAMPLE['share_groups'], _quota_class_set.share_groups
        )
        self.assertEqual(EXAMPLE['gigabytes'], _quota_class_set.gigabytes)
        self.assertEqual(
            EXAMPLE['share_group_snapshots'],
            _quota_class_set.share_group_snapshots,
        )
        self.assertEqual(EXAMPLE['snapshots'], _quota_class_set.snapshots)
        self.assertEqual(
            EXAMPLE['snapshot_gigabytes'], _quota_class_set.snapshot_gigabytes
        )
        self.assertEqual(EXAMPLE['shares'], _quota_class_set.shares)
        self.assertEqual(EXAMPLE['id'], _quota_class_set.id)
        self.assertEqual(
            EXAMPLE['share_networks'], _quota_class_set.share_networks
        )
        self.assertEqual(
            EXAMPLE['share_replicas'], _quota_class_set.share_replicas
        )
        self.assertEqual(
            EXAMPLE['replica_gigabytes'], _quota_class_set.replica_gigabytes
        )
        self.assertEqual(
            EXAMPLE['per_share_gigabytes'],
            _quota_class_set.per_share_gigabytes,
        )
        self.assertEqual(EXAMPLE['backups'], _quota_class_set.backups)
        self.assertEqual(
            EXAMPLE['backup_gigabytes'], _quota_class_set.backup_gigabytes
        )

    def test_update_quota_class_set(self):
        _quota_class_set = quota_class_set.QuotaClassSet(**EXAMPLE)
        updated_attributes = {
            "share_groups": 100,
            "gigabytes": 2000,
            "share_group_snapshots": 100,
        }
        _quota_class_set._update(**updated_attributes)

        self.assertEqual(
            updated_attributes['share_groups'], _quota_class_set.share_groups
        )
        self.assertEqual(
            updated_attributes['gigabytes'], _quota_class_set.gigabytes
        )
        self.assertEqual(
            updated_attributes['share_group_snapshots'],
            _quota_class_set.share_group_snapshots,
        )
        self.assertEqual(EXAMPLE['snapshots'], _quota_class_set.snapshots)



# Copyright 2024 RedHat Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack import resource


class ImageTasks(resource.Resource):
    resources_key = 'tasks'
    base_path = '/images/%(image_id)s/tasks'

    allow_list = True

    _max_microversion = '2.17'

    #: The type of task represented by this content
    type = resource.Body('type')
    #: The current status of this task. The value can be pending, processing,
    #: success or failure
    status = resource.Body('status')
    #: An identifier for the owner of the task, usually the tenant ID
    owner = resource.Body('owner')
    #: The date and time when the task is subject to removal (ISO8601 format)
    expires_at = resource.Body('expires_at')
    #: The date and time when the task was created (ISO8601 format)
    created_at = resource.Body('created_at')
    #: The date and time when the task was updated (ISO8601 format)
    updated_at = resource.Body('updated_at')
    #: The date and time when the task was deleted (ISO8601 format)
    deleted_at = resource.Body('deleted_at')
    #: Whether the task was deleted
    deleted = resource.Body('deleted')
    #: The ID of the image associated to this task
    image_id = resource.Body('image_id')
    #: The request ID of the user message
    request_id = resource.Body('request_id')
    #: The user id associated with this task
    user_id = resource.Body('user_id')
    #: A JSON object specifying the input parameters to the task
    input = resource.Body('input')
    #: A JSON object specifying information about the ultimate outcome of the
    #: task
    result = resource.Body('result')
    #: Human-readable text, possibly an empty string, usually displayed in a
    #: error situation to provide more information about what has occurred
    message = resource.Body('message')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.key_manager.v1 import secret
from openstack.tests.unit import base

ID_VAL = "123"
IDENTIFIER = f'http://localhost:9311/v1/secrets/{ID_VAL}'
EXAMPLE = {
    'algorithm': '1',
    'bit_length': '2',
    'content_types': {'default': '3'},
    'expiration': '2017-03-09T12:14:57.233772',
    'mode': '5',
    'name': '6',
    'secret_ref': IDENTIFIER,
    'status': '8',
    'updated': '2015-03-09T12:15:57.233773',
    'created': '2015-03-09T12:15:57.233774',
    'secret_type': '9',
    'payload': '10',
    'payload_content_type': '11',
    'payload_content_encoding': '12',
}


class TestSecret(base.TestCase):
    def test_basic(self):
        sot = secret.Secret()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('secrets', sot.resources_key)
        self.assertEqual('/secrets', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                "name": "name",
                "mode": "mode",
                "bits": "bits",
                "secret_type": "secret_type",
                "acl_only": "acl_only",
                "created": "created",
                "updated": "updated",
                "expiration": "expiration",
                "sort": "sort",
                "algorithm": "alg",
                "limit": "limit",
                "marker": "marker",
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = secret.Secret(**EXAMPLE)
        self.assertEqual(EXAMPLE['algorithm'], sot.algorithm)
        self.assertEqual(EXAMPLE['bit_length'], sot.bit_length)
        self.assertEqual(EXAMPLE['content_types'], sot.content_types)
        self.assertEqual(EXAMPLE['expiration'], sot.expires_at)
        self.assertEqual(EXAMPLE['mode'], sot.mode)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['secret_ref'], sot.secret_ref)
        self.assertEqual(EXAMPLE['secret_ref'], sot.id)
        self.assertEqual(ID_VAL, sot.secret_id)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['updated'], sot.updated_at)
        self.assertEqual(EXAMPLE['secret_type'], sot.secret_type)
        self.assertEqual(EXAMPLE['payload'], sot.payload)
        self.assertEqual(
            EXAMPLE['payload_content_type'], sot.payload_content_type
        )
        self.assertEqual(
            EXAMPLE['payload_content_encoding'], sot.payload_content_encoding
        )

    def test_get_no_payload(self):
        sot = secret.Secret(id="id")

        sess = mock.Mock()
        rv = mock.Mock()
        return_body = {"status": "cool"}
        rv.json = mock.Mock(return_value=return_body)
        sess.get = mock.Mock(return_value=rv)

        sot.fetch(sess)

        sess.get.assert_called_once_with("secrets/id")

    def _test_payload(self, sot, metadata, content_type="some/type"):
        metadata_response = mock.Mock()
        # Use copy because the dict gets consumed.
        metadata_response.json = mock.Mock(return_value=metadata.copy())

        payload_response = mock.Mock()
        payload = b"secret info"
        payload_response.content = payload

        sess = mock.Mock()
        sess.get = mock.Mock(side_effect=[metadata_response, payload_response])

        rv = sot.fetch(sess)

        sess.get.assert_has_calls(
            [
                mock.call(
                    "secrets/id",
                ),
                mock.call(
                    "secrets/id/payload",
                    headers={"Accept": content_type},
                    skip_cache=False,
                ),
            ]
        )

        if content_type == "text/plain":
            expected_payload = payload.decode("utf-8")
        else:
            expected_payload = payload
        self.assertEqual(rv.payload, expected_payload)
        self.assertEqual(rv.status, metadata["status"])

    def test_get_with_payload_from_argument(self):
        metadata = {"status": "great"}
        content_type = "some/type"
        sot = secret.Secret(id="id", payload_content_type=content_type)
        self._test_payload(sot, metadata, content_type)

    def test_get_with_payload_from_content_types(self):
        content_type = "some/type"
        metadata = {
            "status": "fine",
            "content_types": {"default": content_type},
        }
        sot = secret.Secret(id="id")
        self._test_payload(sot, metadata, content_type)

    def test_get_with_text_payload(self):
        content_type = "text/plain"
        metadata = {
            "status": "fine",
            "content_types": {"default": content_type},
        }
        sot = secret.Secret(id="id")
        self._test_payload(sot, metadata, content_type)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import logging
import sys
from unittest import mock

import fixtures
import os_service_types
import testtools

import openstack
from openstack import exceptions
from openstack.tests.unit import base
from openstack import utils


class Test_enable_logging(base.TestCase):
    def setUp(self):
        super().setUp()
        self.openstack_logger = mock.Mock()
        self.openstack_logger.handlers = []
        self.ksa_logger_root = mock.Mock()
        self.ksa_logger_root.handlers = []
        self.ksa_logger_1 = mock.Mock()
        self.ksa_logger_1.handlers = []
        self.ksa_logger_2 = mock.Mock()
        self.ksa_logger_2.handlers = []
        self.ksa_logger_3 = mock.Mock()
        self.ksa_logger_3.handlers = []
        self.urllib3_logger = mock.Mock()
        self.urllib3_logger.handlers = []
        self.stevedore_logger = mock.Mock()
        self.stevedore_logger.handlers = []
        self.fake_get_logger = mock.Mock()
        self.fake_get_logger.side_effect = [
            self.openstack_logger,
            self.ksa_logger_root,
            self.urllib3_logger,
            self.stevedore_logger,
            self.ksa_logger_1,
            self.ksa_logger_2,
            self.ksa_logger_3,
        ]
        self.useFixture(
            fixtures.MonkeyPatch('logging.getLogger', self.fake_get_logger)
        )

    def _console_tests(self, level, debug, stream):
        openstack.enable_logging(debug=debug, stream=stream)

        self.assertEqual(self.openstack_logger.addHandler.call_count, 1)
        self.openstack_logger.setLevel.assert_called_with(level)

    def _file_tests(self, level, debug):
        file_handler = mock.Mock()
        self.useFixture(
            fixtures.MonkeyPatch('logging.FileHandler', file_handler)
        )
        fake_path = "fake/path.log"

        openstack.enable_logging(debug=debug, path=fake_path)

        file_handler.assert_called_with(fake_path)
        self.assertEqual(self.openstack_logger.addHandler.call_count, 1)
        self.openstack_logger.setLevel.assert_called_with(level)

    def test_none(self):
        openstack.enable_logging(debug=True)
        self.fake_get_logger.assert_has_calls([])
        self.openstack_logger.setLevel.assert_called_with(logging.DEBUG)
        self.assertEqual(self.openstack_logger.addHandler.call_count, 1)
        self.assertIsInstance(
            self.openstack_logger.addHandler.call_args_list[0][0][0],
            logging.StreamHandler,
        )

    def test_debug_console_stderr(self):
        self._console_tests(logging.DEBUG, True, sys.stderr)

    def test_warning_console_stderr(self):
        self._console_tests(logging.INFO, False, sys.stderr)

    def test_debug_console_stdout(self):
        self._console_tests(logging.DEBUG, True, sys.stdout)

    def test_warning_console_stdout(self):
        self._console_tests(logging.INFO, False, sys.stdout)

    def test_debug_file(self):
        self._file_tests(logging.DEBUG, True)

    def test_warning_file(self):
        self._file_tests(logging.INFO, False)


class Test_urljoin(base.TestCase):
    def test_strings(self):
        root = "http://www.example.com"
        leaves = "foo", "bar"

        result = utils.urljoin(root, *leaves)
        self.assertEqual(result, "http://www.example.com/foo/bar")

    def test_with_none(self):
        root = "http://www.example.com"
        leaves = "foo", None

        result = utils.urljoin(root, *leaves)
        self.assertEqual(result, "http://www.example.com/foo/")

    def test_unicode_strings(self):
        root = "http://www.example.com"
        leaves = "ascii", "extra_chars-™"

        try:
            result = utils.urljoin(root, *leaves)
        except Exception:
            self.fail("urljoin failed on unicode strings")

        self.assertEqual(result, "http://www.example.com/ascii/extra_chars-™")


class TestSupportsMicroversion(base.TestCase):
    def setUp(self):
        super().setUp()
        self.adapter = mock.Mock(spec=['get_endpoint_data'])
        self.endpoint_data = mock.Mock(
            spec=['min_microversion', 'max_microversion'],
            min_microversion='1.1',
            max_microversion='1.99',
        )
        self.adapter.get_endpoint_data.return_value = self.endpoint_data

    def test_requested_supported_no_default(self):
        self.adapter.default_microversion = None
        self.assertTrue(utils.supports_microversion(self.adapter, '1.2'))

    def test_requested_not_supported_no_default(self):
        self.adapter.default_microversion = None
        self.assertFalse(utils.supports_microversion(self.adapter, '2.2'))

    def test_requested_not_supported_no_default_exception(self):
        self.adapter.default_microversion = None
        self.assertRaises(
            exceptions.SDKException,
            utils.supports_microversion,
            self.adapter,
            '2.2',
            True,
        )

    def test_requested_supported_higher_default(self):
        self.adapter.default_microversion = '1.8'
        self.assertTrue(utils.supports_microversion(self.adapter, '1.6'))

    def test_requested_supported_equal_default(self):
        self.adapter.default_microversion = '1.8'
        self.assertTrue(utils.supports_microversion(self.adapter, '1.8'))

    def test_requested_supported_lower_default(self):
        self.adapter.default_microversion = '1.2'
        self.assertFalse(utils.supports_microversion(self.adapter, '1.8'))

    def test_requested_supported_lower_default_exception(self):
        self.adapter.default_microversion = '1.2'
        self.assertRaises(
            exceptions.SDKException,
            utils.supports_microversion,
            self.adapter,
            '1.8',
            True,
        )

    @mock.patch('openstack.utils.supports_microversion')
    def test_require_microversion(self, sm_mock):
        utils.require_microversion(self.adapter, '1.2')
        sm_mock.assert_called_with(self.adapter, '1.2', raise_exception=True)


class TestMaximumSupportedMicroversion(base.TestCase):
    def setUp(self):
        super().setUp()
        self.adapter = mock.Mock(spec=['get_endpoint_data'])
        self.endpoint_data = mock.Mock(
            spec=['min_microversion', 'max_microversion'],
            min_microversion=None,
            max_microversion='1.99',
        )
        self.adapter.get_endpoint_data.return_value = self.endpoint_data

    def test_with_none(self):
        self.assertIsNone(
            utils.maximum_supported_microversion(self.adapter, None)
        )

    def test_with_value(self):
        self.assertEqual(
            '1.42', utils.maximum_supported_microversion(self.adapter, '1.42')
        )

    def test_value_more_than_max(self):
        self.assertEqual(
            '1.99', utils.maximum_supported_microversion(self.adapter, '1.100')
        )

    def test_value_less_than_min(self):
        self.endpoint_data.min_microversion = '1.42'
        self.assertIsNone(
            utils.maximum_supported_microversion(self.adapter, '1.2')
        )


class TestOsServiceTypesVersion(base.TestCase):
    def test_ost_version(self):
        ost_version = '2022-09-13T15:34:32.154125'
        self.assertEqual(
            ost_version,
            os_service_types.ServiceTypes().version,
            "This project must be pinned to the latest version of "
            "os-service-types. Please bump requirements.txt accordingly.",
        )


class TestTinyDAG(base.TestCase):
    test_graph = {
        'a': ['b', 'd', 'f'],
        'b': ['c', 'd'],
        'c': ['d'],
        'd': ['e'],
        'e': [],
        'f': ['e'],
        'g': ['e'],
    }

    @classmethod
    def _create_tinydag(cls, data):
        sot = utils.TinyDAG()
        for k, v in data.items():
            sot.add_node(k)
            for dep in v:
                sot.add_edge(k, dep)
        return sot

    def _verify_order(self, test_graph, test_list):
        for k, v in test_graph.items():
            for dep in v:
                self.assertTrue(test_list.index(k) < test_list.index(dep))

    def test_topological_sort(self):
        sot = self._create_tinydag(self.test_graph)
        sorted_list = sot.topological_sort()
        self._verify_order(sot.graph, sorted_list)
        self.assertEqual(len(self.test_graph.keys()), len(sorted_list))

    def test_walk(self):
        sot = self._create_tinydag(self.test_graph)
        sorted_list = []
        for node in sot.walk():
            sorted_list.append(node)
            sot.node_done(node)
        self._verify_order(sot.graph, sorted_list)
        self.assertEqual(len(self.test_graph.keys()), len(sorted_list))

    def test_walk_parallel(self):
        sot = self._create_tinydag(self.test_graph)
        sorted_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=15) as executor:
            for node in sot.walk(timeout=1):
                executor.submit(test_walker_fn, sot, node, sorted_list)
        self._verify_order(sot.graph, sorted_list)
        self.assertEqual(len(self.test_graph.keys()), len(sorted_list))

    def test_walk_raise(self):
        sot = self._create_tinydag(self.test_graph)
        bad_node = 'f'
        with testtools.ExpectedException(exceptions.SDKException):
            for node in sot.walk(timeout=1):
                if node != bad_node:
                    sot.node_done(node)

    def test_add_node_after_edge(self):
        sot = utils.TinyDAG()
        sot.add_node('a')
        sot.add_edge('a', 'b')
        sot.add_node('a')
        self.assertEqual(sot._graph['a'], set('b'))


def test_walker_fn(graph, node, lst):
    lst.append(node)
    graph.node_done(node)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.baremetal.v1 import _common
from openstack import resource


class Port(_common.Resource):
    resources_key = 'ports'
    base_path = '/ports'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_patch = True
    commit_method = 'PATCH'
    commit_jsonpatch = True

    _query_mapping = resource.QueryParameters(
        'address',
        'node',
        'portgroup',
        'shard',
        fields={'type': _common.fields_type},
        node_id='node_uuid',
    )

    # The physical_network field introduced in 1.34
    # The is_smartnic field added in 1.53
    # Query filter by shard added in 1.82
    # The name field added in 1.88
    # The description field added in 1.97
    _max_microversion = '1.97'

    #: The physical hardware address of the network port, typically the
    #: hardware MAC address.
    address = resource.Body('address')
    #: Timestamp at which the port was created.
    created_at = resource.Body('created_at')
    #: The description for the port
    description = resource.Body('description')
    #: A set of one or more arbitrary metadata key and value pairs.
    extra = resource.Body('extra')
    #: The UUID of the port
    id = resource.Body('uuid', alternate_id=True)
    #: Internal metadata set and stored by the port. This field is read-only.
    #: Added in API microversion 1.18.
    internal_info = resource.Body('internal_info')
    #: Whether PXE is enabled on the port. Added in API microversion 1.19.
    is_pxe_enabled = resource.Body('pxe_enabled', type=bool)
    #: Whether the port is a Smart NIC port. Added in API microversion 1.53.
    is_smartnic = resource.Body('is_smartnic', type=bool)
    #: A list of relative links, including the self and bookmark links.
    links = resource.Body('links', type=list)
    #: The port bindig profile. If specified, must contain ``switch_id`` and
    #: ``port_id`` fields. ``switch_info`` field is an optional string field
    #: to be used to store vendor specific information. Added in API
    #: microversion 1.19.
    local_link_connection = resource.Body('local_link_connection')
    #: The name of the port
    name = resource.Body('name')
    #: The UUID of node this port belongs to
    node_id = resource.Body('node_uuid')
    #: The name of physical network this port is attached to.
    #: Added in API microversion 1.34.
    physical_network = resource.Body('physical_network')
    #: The UUID of PortGroup this port belongs to. Added in API microversion
    #: 1.24.
    port_group_id = resource.Body('portgroup_uuid')
    #: Read-only. The parent port trunk details dictionary, with the trunk ID
    # and the subports information (port ID, segmentation ID and segmentation
    # type).
    trunk_details = resource.Body('trunk_details', type=dict)
    #: Timestamp at which the port was last updated.
    updated_at = resource.Body('updated_at')


PortDetail = Port



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import _base
from openstack import resource


class SecurityGroup(_base.NetworkResource, _base.TagMixinNetwork):
    resource_key = 'security_group'
    resources_key = 'security_groups'
    base_path = '/security-groups'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description',
        'fields',
        'id',
        'name',
        'stateful',
        'project_id',
        'tenant_id',
        'revision_number',
        'sort_dir',
        'sort_key',
        is_shared='shared',
        **_base.TagMixinNetwork._tag_query_parameters,
    )

    # Properties
    #: Timestamp when the security group was created.
    created_at = resource.Body('created_at')
    #: The security group description.
    description = resource.Body('description')
    #: The security group name.
    name = resource.Body('name')
    #: Whether the security group is stateful or not.
    stateful = resource.Body('stateful')
    #: The ID of the project this security group is associated with.
    project_id = resource.Body('project_id')
    #: A list of
    #: :class:`~openstack.network.v2.security_group_rule.SecurityGroupRule`
    #: objects. *Type: list*
    security_group_rules = resource.Body('security_group_rules', type=list)
    #: The ID of the project this security group is associated with.
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: Timestamp when the security group was last updated.
    updated_at = resource.Body('updated_at')
    #: Indicates whether this Security Group is shared across all projects.
    #: *Type: bool*
    is_shared = resource.Body('shared', type=bool)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
The :class:`~openstack.connection.Connection` class is the primary interface
to the Python SDK. It maintains a context for a connection to a region of
a cloud provider. The :class:`~openstack.connection.Connection` has an
attribute to access each OpenStack service.

At a minimum, the :class:`~openstack.connection.Connection` class needs to be
created with a config or the parameters to build one.

While the overall system is very flexible, there are four main use cases
for different ways to create a :class:`~openstack.connection.Connection`.

* Using config settings and keyword arguments as described in
  :ref:`openstack-config`
* Using only keyword arguments passed to the constructor ignoring config files
  and environment variables.
* Using an existing authenticated `keystoneauth1.session.Session`, such as
  might exist inside of an OpenStack service operational context.
* Using an existing :class:`~openstack.config.cloud_region.CloudRegion`.

Creating the Connection
-----------------------

Using config settings
~~~~~~~~~~~~~~~~~~~~~

For users who want to create a :class:`~openstack.connection.Connection` making
use of named clouds in ``clouds.yaml`` files, ``OS_`` environment variables
and python keyword arguments, the :func:`openstack.connect` factory function
is the recommended way to go:

.. code-block:: python

    import openstack

    conn = openstack.connect(cloud='example', region_name='earth1')

If the application in question is a command line application that should also
accept command line arguments, an `argparse.Namespace` can be passed to
:func:`openstack.connect` that will have relevant arguments added to it and
then subsequently consumed by the constructor:

.. code-block:: python

    import argparse
    import openstack

    options = argparse.ArgumentParser(description='Awesome OpenStack App')
    conn = openstack.connect(options=options)

Using only keyword arguments
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If the application wants to avoid loading any settings from ``clouds.yaml`` or
environment variables, use the :class:`~openstack.connection.Connection`
constructor directly. As long as the ``cloud`` argument is omitted or ``None``,
the :class:`~openstack.connection.Connection` constructor will not load
settings from files or the environment.

.. note::

    This is a different default behavior than the :func:`~openstack.connect`
    factory function. In :func:`~openstack.connect` if ``cloud`` is omitted
    or ``None``, a default cloud will be loaded, defaulting to the ``envvars``
    cloud if it exists.

.. code-block:: python

    from openstack import connection

    conn = connection.Connection(
        region_name='example-region',
        auth={
            'auth_url': 'https://auth.example.com',
            'username': 'amazing-user',
            'password': 'super-secret-password',
            'project_id': '33aa1afc-03fe-43b8-8201-4e0d3b4b8ab5',
            'user_domain_id': '054abd68-9ad9-418b-96d3-3437bb376703',
        },
        compute_api_version='2',
        identity_interface='internal',
    )

Per-service settings as needed by `keystoneauth1.adapter.Adapter` such as
``api_version``, ``service_name``, and ``interface`` can be set, as seen
above, by prefixing them with the official ``service-type`` name of the
service. ``region_name`` is a setting for the entire
:class:`~openstack.config.cloud_region.CloudRegion` and cannot be set per
service.

From existing authenticated Session
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

For applications that already have an authenticated Session, simply passing
it to the :class:`~openstack.connection.Connection` constructor is all that
is needed:

.. code-block:: python

    from openstack import connection

    conn = connection.Connection(
        session=session,
        region_name='example-region',
        compute_api_version='2',
        identity_interface='internal',
    )

From oslo.conf CONF object
~~~~~~~~~~~~~~~~~~~~~~~~~~

For applications that have an oslo.config ``CONF`` object that has been
populated with ``keystoneauth1.loading.register_adapter_conf_options`` in
groups named by the OpenStack service's project name, it is possible to
construct a Connection with the ``CONF`` object and an authenticated Session.

.. note::

    This is primarily intended for use by OpenStack services to talk amongst
    themselves.

.. code-block:: python

    from keystoneauth1 import loading as ks_loading
    from openstack import connection
    from oslo_config import cfg

    CONF = cfg.CONF

    group = cfg.OptGroup('neutron')
    ks_loading.register_session_conf_options(CONF, group)
    ks_loading.register_auth_conf_options(CONF, group)
    ks_loading.register_adapter_conf_options(CONF, group)

    CONF()

    auth = ks_loading.load_auth_from_conf_options(CONF, 'neutron')
    sess = ks_loading.load_session_from_conf_options(CONF, 'neutron', auth=auth)

    conn = connection.Connection(
        session=sess,
        oslo_conf=CONF,
    )

This can then be used with an appropriate configuration file.

.. code-block:: ini

    [neutron]
    region_name = RegionOne
    auth_strategy = keystone
    project_domain_name = Default
    project_name = service
    user_domain_name = Default
    password = password
    username = neutron
    auth_url = http://10.0.110.85/identity
    auth_type = password
    service_metadata_proxy = True
    default_floating_pool = public

You may also wish to configure a service user. As discussed in the `Keystone
documentation`__, service users are users with specific roles that identify the
user as a service. The use of service users can avoid issues caused by the
expiration of the original user's token during long running operations, as a
fresh token issued for the service user will always accompany the user's token,
which may have expired.

.. code-block:: python

    from keystoneauth1 import loading as ks_loading
    from keystoneauth1 import service_token
    from oslo_config import cfg
    import openstack
    from openstack import connection

    CONF = cfg.CONF

    neutron_group = cfg.OptGroup('neutron')
    ks_loading.register_session_conf_options(CONF, neutron_group)
    ks_loading.register_auth_conf_options(CONF, neutron_group)
    ks_loading.register_adapter_conf_options(CONF, neutron_group)

    service_group = cfg.OptGroup('service_user')
    ks_loading.register_session_conf_options(CONF, service_group)
    ks_loading.register_auth_conf_options(CONF, service_group)

    CONF()
    user_auth = ks_loading.load_auth_from_conf_options(CONF, 'neutron')
    service_auth = ks_loading.load_auth_from_conf_options(CONF, 'service_user')
    auth = service_token.ServiceTokenAuthWrapper(user_auth, service_auth)

    sess = ks_loading.load_session_from_conf_options(CONF, 'neutron', auth=auth)

    conn = connection.Connection(
        session=sess,
        oslo_conf=CONF,
    )

This will necessitate an additional section in the configuration file used.

.. code-block:: ini

    [service_user]
    auth_strategy = keystone
    project_domain_name = Default
    project_name = service
    user_domain_name = Default
    password = password
    username = nova
    auth_url = http://10.0.110.85/identity
    auth_type = password

.. __: https://docs.openstack.org/keystone/latest/admin/manage-services.html

From existing CloudRegion
~~~~~~~~~~~~~~~~~~~~~~~~~

If you already have an :class:`~openstack.config.cloud_region.CloudRegion`
you can pass it in instead:

.. code-block:: python

    from openstack import connection
    import openstack.config

    config = openstack.config.get_cloud_region(
        cloud='example',
        region_name='earth',
    )
    conn = connection.Connection(config=config)

Using the Connection
--------------------

Services are accessed through an attribute named after the service's official
service-type.

List
~~~~

An iterator containing a list of all the projects is retrieved in this manner:

.. code-block:: python

    projects = conn.identity.projects()

Find or create
~~~~~~~~~~~~~~

If you wanted to make sure you had a network named 'zuul', you would first
try to find it and if that fails, you would create it::

    network = conn.network.find_network("zuul")
    if network is None:
        network = conn.network.create_network(name="zuul")

Additional information about the services can be found in the
:ref:`service-proxies` documentation.
"""

import argparse
import concurrent.futures
import copy
import importlib.metadata as importlib_metadata
import typing as ty
import warnings

import keystoneauth1.exceptions
from keystoneauth1 import session as ks_session
import requestsexceptions
import typing_extensions as ty_ext

from openstack import _log
from openstack.cloud import _accelerator
from openstack.cloud import _baremetal
from openstack.cloud import _block_storage
from openstack.cloud import _coe
from openstack.cloud import _compute
from openstack.cloud import _dns
from openstack.cloud import _identity
from openstack.cloud import _image
from openstack.cloud import _network
from openstack.cloud import _object_store
from openstack.cloud import _orchestration
from openstack.cloud import _shared_file_system
from openstack import config as _config
import openstack.config.cloud_region
from openstack import exceptions
from openstack import service_description

if ty.TYPE_CHECKING:
    from oslo_config import cfg

    from openstack.config import cloud_region
    from openstack import proxy

__all__ = [
    'from_config',
    'Connection',
]

if requestsexceptions.SubjectAltNameWarning:
    warnings.filterwarnings(
        'ignore', category=requestsexceptions.SubjectAltNameWarning
    )

_logger = _log.setup_logging('openstack')


def from_config(
    cloud: str | None = None,
    config: ty.Optional['cloud_region.CloudRegion'] = None,
    options: argparse.Namespace | None = None,
    **kwargs: ty.Any,
) -> 'Connection':
    """Create a Connection using openstack.config

    :param str cloud:
        Use the `cloud` configuration details when creating the Connection.
    :param openstack.config.cloud_region.CloudRegion config:
        An existing CloudRegion configuration. If no `config` is provided,
        `openstack.config.OpenStackConfig` will be called, and the provided
        `name` will be used in determining which cloud's configuration
        details will be used in creation of the `Connection` instance.
    :param argparse.Namespace options:
        Allows direct passing in of options to be added to the cloud config.
        This does not have to be an actual instance of argparse.Namespace,
        despite the naming of the
        `openstack.config.loader.OpenStackConfig.get_one` argument to which
        it is passed.

    :rtype: :class:`~openstack.connection.Connection`
    """
    # TODO(mordred) Backwards compat while we transition
    cloud = kwargs.pop('cloud_name', cloud)
    config = kwargs.pop('cloud_config', config)
    if config is None:
        config = _config.OpenStackConfig().get_one(
            cloud=cloud, argparse=options, **kwargs
        )

    return Connection(config=config)


class Connection(
    _accelerator.AcceleratorCloudMixin,
    _baremetal.BaremetalCloudMixin,
    _block_storage.BlockStorageCloudMixin,
    _compute.ComputeCloudMixin,
    _coe.CoeCloudMixin,
    _dns.DnsCloudMixin,
    _identity.IdentityCloudMixin,
    _image.ImageCloudMixin,
    _network.NetworkCloudMixin,
    _object_store.ObjectStoreCloudMixin,
    _orchestration.OrchestrationCloudMixin,
    _shared_file_system.SharedFileSystemCloudMixin,
):
    def __init__(
        self,
        cloud: str | None = None,
        config: ty.Optional['cloud_region.CloudRegion'] = None,
        session: ks_session.Session | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
        extra_services: list[service_description.ServiceDescription]
        | None = None,
        strict: bool = False,
        use_direct_get: bool | None = None,
        task_manager: ty.Any = None,
        rate_limit: float | dict[str, float] | None = None,
        oslo_conf: ty.Optional['cfg.ConfigOpts'] = None,
        service_types: list[str] | None = None,
        global_request_id: str | None = None,
        strict_proxies: bool = False,
        pool_executor: concurrent.futures.Executor | None = None,
        **kwargs: ty.Any,
    ):
        """Create a connection to a cloud.

        A connection needs information about how to connect, how to
        authenticate and how to select the appropriate services to use.

        The recommended way to provide this information is by referencing
        a named cloud config from an existing `clouds.yaml` file. The cloud
        name ``envvars`` may be used to consume a cloud configured via ``OS_``
        environment variables.

        A pre-existing :class:`~openstack.config.cloud_region.CloudRegion`
        object can be passed in lieu of a cloud name, for cases where the user
        already has a fully formed CloudRegion and just wants to use it.

        Similarly, if for some reason the user already has a
        :class:`~keystoneauth1.session.Session` and wants to use it, it may be
        passed in.

        :param str cloud: Name of the cloud from config to use.
        :param config: CloudRegion object representing the config for the
            region of the cloud in question.
        :type config: :class:`~openstack.config.cloud_region.CloudRegion`
        :param session: A session object compatible with
            :class:`~keystoneauth1.session.Session`.
        :type session: :class:`~keystoneauth1.session.Session`
        :param str app_name: Name of the application to be added to User Agent.
        :param str app_version: Version of the application to be added to
            User Agent.
        :param extra_services: List of
            :class:`~openstack.service_description.ServiceDescription`
            objects describing services that openstacksdk otherwise does not
            know about.
        :param bool use_direct_get:
            For get methods, make specific REST calls for server-side
            filtering instead of making list calls and filtering client-side.
            Default false.
        :param task_manager:
            Ignored. Exists for backwards compat during transition. Rate limit
            parameters should be passed directly to the `rate_limit` parameter.
        :param rate_limit:
            Client-side rate limit, expressed in calls per second. The
            parameter can either be a single float, or it can be a dict with
            keys as service-type and values as floats expressing the calls
            per second for that service. Defaults to None, which means no
            rate-limiting is performed.
        :param oslo_conf: An oslo.config ``CONF`` object that has been
            populated with
            ``keystoneauth1.loading.register_adapter_conf_options`` in
            groups named by the OpenStack service's project name.
        :type oslo_conf: :class:`~oslo_config.cfg.ConfigOpts`
        :param service_types:
            A list/set of service types this Connection should support. All
            other service types will be disabled (will error if used).
            **Currently only supported in conjunction with the ``oslo_conf``
            kwarg.**
        :param strict_proxies:
            Throw an ``openstack.exceptions.ServiceDiscoveryException`` if the
            endpoint for a given service doesn't work. This is useful for
            OpenStack services using sdk to talk to other OpenStack services
            where it can be expected that the deployer config is correct and
            errors should be reported immediately.
            Default false.
        :type strict_proxies: bool
        :param global_request_id: A Request-id to send with all interactions.
        :type global_request_id: str
        :param pool_executor:
            A futurist ``Executor`` object to be used for concurrent background
            activities. Defaults to None in which case a ThreadPoolExecutor
            will be created if needed.
        :type pool_executor: :class:`~futurist.Executor`
        :param kwargs: If a config is not provided, the rest of the parameters
            provided are assumed to be arguments to be passed to the
            CloudRegion constructor.
        """
        super().__init__(
            cloud=cloud,
            config=config,
            session=session,
            app_name=app_name,
            app_version=app_version,
            extra_services=extra_services,
            strict=strict,
            use_direct_get=use_direct_get,
            task_manager=task_manager,
            rate_limit=rate_limit,
            oslo_conf=oslo_conf,
            service_types=service_types,
            global_request_id=global_request_id,
            strict_proxies=strict_proxies,
            pool_executor=pool_executor,
            **kwargs,
        )

        # Allow vendors to provide hooks. They will normally only receive a
        # connection object and a responsible to register additional services
        vendor_hook = kwargs.get('vendor_hook')
        if not vendor_hook and 'vendor_hook' in self.config.config:
            # Get the one from profile
            vendor_hook = self.config.config.get('vendor_hook')
        if vendor_hook:
            try:
                # NOTE(gtema): no class name in the hook, plain module:function
                # Split string hook into module and function
                try:
                    package_name, function = vendor_hook.rsplit(':')

                    if package_name and function:
                        ep = importlib_metadata.EntryPoint(
                            name='vendor_hook',
                            value=vendor_hook,
                            group='vendor_hook',
                        )
                        hook = ep.load()
                        hook(self)
                except ValueError:
                    self.log.warning(
                        'Hook should be in the entrypoint '
                        'module:attribute format'
                    )
            except (ImportError, TypeError, AttributeError) as e:
                self.log.warning(
                    'Configured hook %s cannot be executed: %s', vendor_hook, e
                )

        # Add additional metrics into the configuration according to the
        # selected connection. We don't want to deal with overall config in the
        # proxy, just pass required part.
        if (
            self.config._influxdb_config
            and 'additional_metric_tags' in self.config.config
        ):
            self.config._influxdb_config['additional_metric_tags'] = (
                self.config.config['additional_metric_tags']
            )

    def add_service(
        self, service: service_description.ServiceDescription
    ) -> None:
        """Add a service to the Connection.

        Attaches an instance of the :class:`~openstack.proxy.Proxy`
        class contained in
        :class:`~openstack.service_description.ServiceDescription`.
        The :class:`~openstack.proxy.Proxy` will be attached to the
        `Connection` by its ``service_type`` and by any ``aliases`` that
        may be specified.

        :param openstack.service_description.ServiceDescription service:
            Object describing the service to be attached. As a convenience,
            if ``service`` is a string it will be treated as a ``service_type``
            and a basic
            :class:`~openstack.service_description.ServiceDescription`
            will be created.
        """
        # If we don't have a proxy, just instantiate Proxy so that
        # we get an adapter.
        if isinstance(service, str):
            service = service_description.ServiceDescription(service)

        # Directly invoke descriptor of the ServiceDescription
        def getter(self: 'Connection') -> 'proxy.Proxy':
            # TODO(stephenfin): Remove ignore once we have typed
            # ServiceDescription
            return service.__get__(self, service)  # type: ignore

        # Register the ServiceDescription class (as property)
        # with every known alias for a "runtime descriptor"
        for attr_name in service.all_types:
            setattr(
                self.__class__,
                attr_name.replace('-', '_'),
                property(fget=getter),
            )
        self.config.enable_service(service.service_type)

    def authorize(self) -> str:
        """Authorize this Connection

        .. note::

            This method is optional. When an application makes a call to any
            OpenStack service, this method allows you to request a token
            manually before attempting to do anything else.

        :returns: A string token.
        :raises: :class:`~openstack.exceptions.HttpException` if the
            authorization fails due to reasons like the credentials provided
            are unable to be authorized or the `auth_type` argument is missing,
            etc.
        """
        try:
            return ty.cast(str, self.session.get_token())
        except keystoneauth1.exceptions.ClientException as e:
            raise exceptions.SDKException(str(e))

    def connect_as(self, **kwargs: ty.Any) -> ty_ext.Self:
        """Make a new Connection object with new auth context.

        Take the existing settings from the current cloud and construct a new
        Connection object with some of the auth settings overridden. This
        is useful for getting an object to perform tasks with as another user,
        or in the context of a different project.

        .. code-block:: python

            conn = openstack.connect(cloud='example')
            # Work normally
            servers = conn.list_servers()
            conn2 = conn.connect_as(username='different-user', password='')
            # Work as different-user
            servers = conn2.list_servers()

        :param kwargs: keyword arguments can contain anything that would
            normally go in an auth dict. They will override the same settings
            from the parent cloud as appropriate. Entries that do not want to
            be overridden can be ommitted.
        """

        if self.config._openstack_config:
            config = self.config._openstack_config
        else:
            # TODO(mordred) Replace this with from_session
            config = openstack.config.OpenStackConfig(
                app_name=self.config._app_name,
                app_version=self.config._app_version,
                load_yaml_config=False,
            )
        params = copy.deepcopy(self.config.config)
        # Remove profile from current cloud so that overridding works
        params.pop('profile', None)

        # Utility function to help with the stripping below.
        def pop_keys(
            params: dict[str, dict[str, str | None]],
            auth: dict[str, str | None],
            name_key: str,
            id_key: str,
        ) -> None:
            if name_key in auth or id_key in auth:
                params['auth'].pop(name_key, None)
                params['auth'].pop(id_key, None)

        # If there are user, project or domain settings in the incoming auth
        # dict, strip out both id and name so that a user can say:
        #     cloud.connect_as(project_name='foo')
        # and have that work with clouds that have a project_id set in their
        # config.
        for prefix in ('user', 'project'):
            if prefix == 'user':
                name_key = 'username'
            else:
                name_key = 'project_name'
            id_key = f'{prefix}_id'
            pop_keys(params, kwargs, name_key, id_key)
            id_key = f'{prefix}_domain_id'
            name_key = f'{prefix}_domain_name'
            pop_keys(params, kwargs, name_key, id_key)

        for key, value in kwargs.items():
            params['auth'][key] = value

        cloud_region = config.get_one(**params)
        # Attach the discovery cache from the old session so we won't
        # double discover.
        cloud_region._discovery_cache = self.session._discovery_cache
        # Override the cloud name so that logging/location work right
        cloud_region._name = self.name
        cloud_region.config['profile'] = self.name
        # Use self.__class__ so that we return whatever this if, like if it's
        # a subclass in the case of shade wrapping sdk.
        return self.__class__(config=cloud_region)

    def connect_as_project(self, project: str) -> ty_ext.Self:
        """Make a new Connection object with a new project.

        Take the existing settings from the current cloud and construct a new
        Connection object with the project settings overridden. This
        is useful for getting an object to perform tasks with as another user,
        or in the context of a different project.

        .. code-block:: python

            cloud = openstack.connect(cloud='example')
            # Work normally
            servers = cloud.list_servers()
            cloud2 = cloud.connect_as_project('different-project')
            # Work in different-project
            servers = cloud2.list_servers()

        :param project: Either a project name or a project dict as returned by
            ``list_projects``.
        """
        auth = {}
        if isinstance(project, dict):
            auth['project_id'] = project.get('id')
            auth['project_name'] = project.get('name')
            if project.get('domain_id'):
                auth['project_domain_id'] = project['domain_id']
        else:
            auth['project_name'] = project
        return self.connect_as(**auth)

    def endpoint_for(
        self,
        service_type: str,
        interface: str | None = None,
        region_name: str | None = None,
    ) -> str | None:
        """Return the endpoint for a given service.

        Respects config values for Connection, including
        ``*_endpoint_override``. For direct values from the catalog regardless
        of overrides, see
        :meth:`~openstack.config.cloud_region.CloudRegion.get_endpoint_from_catalog`

        :param service_type: Service Type of the endpoint to search for.
        :param interface: Interface of the endpoint to search for. Optional,
            defaults to the configured value for interface for this Connection.
        :param region_name: Region Name of the endpoint to search for.
            Optional, defaults to the configured value for region_name for this
            Connection.

        :returns: The endpoint of the service, or None if not found.
        """

        endpoint_override = self.config.get_endpoint(service_type)
        if endpoint_override:
            return endpoint_override
        return self.config.get_endpoint_from_catalog(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from examples.compute import create
from examples.compute import delete
from examples.compute import find as compute_find
from examples.compute import list as compute_list
from examples.network import find as network_find
from examples.network import list as network_list

from openstack.tests.functional import base


class TestCompute(base.BaseFunctionalTest):
    """Test the compute examples

    The purpose of these tests is to ensure the examples run successfully.
    """

    def test_compute(self):
        compute_list.list_servers(self.operator_cloud)
        compute_list.list_images(self.operator_cloud)
        compute_list.list_flavors(self.operator_cloud)
        compute_list.list_keypairs(self.operator_cloud)
        network_list.list_networks(self.operator_cloud)

        compute_find.find_image(self.operator_cloud)
        compute_find.find_flavor(self.operator_cloud)
        compute_find.find_keypair(self.operator_cloud)
        network_find.find_network(self.operator_cloud)

        create.create_server(self.operator_cloud)

        delete.delete_keypair(self.operator_cloud)
        delete.delete_server(self.operator_cloud)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.compute.v2 import volume_attachment
from openstack.tests.unit import base


EXAMPLE = {
    'attachment_id': '979ce4f8-033a-409d-85e6-6b5c0f6a6302',
    'delete_on_termination': False,
    'device': '/dev/sdc',
    'serverId': '7696780b-3f53-4688-ab25-019bfcbbd806',
    'tag': 'foo',
    'volumeId': 'a07f71dc-8151-4e7d-a0cc-cd24a3f11113',
    'bdm_uuid': 'c088db45-92b8-49e8-81e2-a1b77a144b3b',
}


class TestServerInterface(base.TestCase):
    def test_basic(self):
        sot = volume_attachment.VolumeAttachment()
        self.assertEqual('volumeAttachment', sot.resource_key)
        self.assertEqual('volumeAttachments', sot.resources_key)
        self.assertEqual(
            '/servers/%(server_id)s/os-volume_attachments',
            sot.base_path,
        )
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertDictEqual(
            {"limit": "limit", "offset": "offset", "marker": "marker"},
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = volume_attachment.VolumeAttachment(**EXAMPLE)
        self.assertEqual(EXAMPLE['volumeId'], sot.id)
        self.assertEqual(EXAMPLE['attachment_id'], sot.attachment_id)
        self.assertEqual(
            EXAMPLE['delete_on_termination'],
            sot.delete_on_termination,
        )
        self.assertEqual(EXAMPLE['device'], sot.device)
        # FIXME(stephenfin): This conflicts since there is a server ID in the
        # URI *and* in the body. We need a field that handles both or we need
        # to use different names.
        # self.assertEqual(EXAMPLE['serverId'], sot.server_id)
        self.assertEqual(EXAMPLE['tag'], sot.tag)
        self.assertEqual(EXAMPLE['volumeId'], sot.volume_id)
        self.assertEqual(EXAMPLE['bdm_uuid'], sot.bdm_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class LoadBalancer(resource.Resource):
    resource_key = 'loadbalancer'
    resources_key = 'loadbalancers'
    base_path = '/lbaas/loadbalancers'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description',
        'name',
        'project_id',
        'provider',
        'provisioning_status',
        'tenant_id',
        'vip_address',
        'vip_subnet_id',
        is_admin_state_up='admin_state_up',
    )

    # Properties
    #: Description for the load balancer.
    description = resource.Body('description')
    #: The administrative state of the load balancer, which is up
    #: ``True`` or down ``False``. *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: List of listeners associated with this load balancer.
    #: *Type: list of dicts which contain the listener IDs*
    listener_ids = resource.Body('listeners', type=list)
    #: Name of the load balancer
    name = resource.Body('name')
    #: Status of load_balancer operating, e.g. ONLINE, OFFLINE.
    operating_status = resource.Body('operating_status')
    #: List of pools associated with this load balancer.
    #: *Type: list of dicts which contain the pool IDs*
    pool_ids = resource.Body('pools', type=list)
    #: The ID of the project this load balancer is associated with.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The name of the provider.
    provider = resource.Body('provider')
    #: Status of load balancer provisioning, e.g. ACTIVE, INACTIVE.
    provisioning_status = resource.Body('provisioning_status')
    #: The IP address of the VIP.
    vip_address = resource.Body('vip_address')
    #: The ID of the port for the VIP.
    vip_port_id = resource.Body('vip_port_id')
    #: The ID of the subnet on which to allocate the VIP address.
    vip_subnet_id = resource.Body('vip_subnet_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.compute.v2 import aggregate
from openstack.tests.unit import base

IDENTIFIER = 'IDENTIFIER'

EXAMPLE = {
    "name": "m-family",
    "availability_zone": None,
    "deleted": False,
    "created_at": "2018-07-06T14:58:16.000000",
    "updated_at": None,
    "hosts": ["oscomp-m001", "oscomp-m002", "oscomp-m003"],
    "deleted_at": None,
    "id": 4,
    "uuid": IDENTIFIER,
    "metadata": {"type": "public", "family": "m-family"},
}


class TestAggregate(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = EXAMPLE.copy()
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.resp.headers = {'Accept': ''}
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.post = mock.Mock(return_value=self.resp)

    def test_basic(self):
        sot = aggregate.Aggregate()
        self.assertEqual('aggregate', sot.resource_key)
        self.assertEqual('aggregates', sot.resources_key)
        self.assertEqual('/os-aggregates', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = aggregate.Aggregate(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['availability_zone'], sot.availability_zone)
        self.assertEqual(EXAMPLE['deleted'], sot.is_deleted)
        self.assertEqual(EXAMPLE['deleted_at'], sot.deleted_at)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['hosts'], sot.hosts)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['uuid'], sot.uuid)
        self.assertDictEqual(EXAMPLE['metadata'], sot.metadata)

    def test_add_host(self):
        sot = aggregate.Aggregate(**EXAMPLE)

        sot.add_host(self.sess, 'host1')

        url = 'os-aggregates/4/action'
        body = {"add_host": {"host": "host1"}}
        self.sess.post.assert_called_with(url, json=body, microversion=None)

    def test_remove_host(self):
        sot = aggregate.Aggregate(**EXAMPLE)

        sot.remove_host(self.sess, 'host1')

        url = 'os-aggregates/4/action'
        body = {"remove_host": {"host": "host1"}}
        self.sess.post.assert_called_with(url, json=body, microversion=None)

    def test_set_metadata(self):
        sot = aggregate.Aggregate(**EXAMPLE)

        sot.set_metadata(self.sess, {"key: value"})

        url = 'os-aggregates/4/action'
        body = {"set_metadata": {"metadata": {"key: value"}}}
        self.sess.post.assert_called_with(url, json=body, microversion=None)

    def test_precache_image(self):
        sot = aggregate.Aggregate(**EXAMPLE)

        sot.precache_images(self.sess, ['1'])

        url = 'os-aggregates/4/images'
        body = {"cache": ['1']}
        self.sess.post.assert_called_with(
            url, json=body, microversion=sot._max_microversion
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import _base
from openstack import resource


class SubnetPool(resource.Resource, _base.TagMixinNetwork):
    resource_key = 'subnetpool'
    resources_key = 'subnetpools'
    base_path = '/subnetpools'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'address_scope_id',
        'description',
        'ip_version',
        'is_default',
        'name',
        'project_id',
        'sort_key',
        'sort_dir',
        is_shared='shared',
        **_base.TagMixinNetwork._tag_query_parameters,
    )

    # Properties
    #: The ID of the address scope associated with the subnet pool.
    address_scope_id = resource.Body('address_scope_id')
    #: Timestamp when the subnet pool was created.
    created_at = resource.Body('created_at')
    #: The length of the prefix to allocate when the cidr or prefixlen
    #: attributes are omitted when creating a subnet. *Type: int*
    default_prefix_length = resource.Body('default_prefixlen', type=int)
    #: A per-project quota on the prefix space that can be allocated
    #: from the subnet pool for project subnets. For IPv4 subnet pools,
    #: default_quota is measured in units of /32. For IPv6 subnet pools,
    #: default_quota is measured units of /64. All projects that use the
    #: subnet pool have the same prefix quota applied. *Type: int*
    default_quota = resource.Body('default_quota', type=int)
    #: The subnet pool description.
    description = resource.Body('description')
    #: Read-only. The IP address family of the list of prefixes.
    #: *Type: int*
    ip_version = resource.Body('ip_version', type=int)
    #: Whether or not this is the default subnet pool.
    #: *Type: bool*
    is_default = resource.Body('is_default', type=bool)
    #: Indicates whether this subnet pool is shared across all projects.
    #: *Type: bool*
    is_shared = resource.Body('shared', type=bool)
    #: The maximum prefix length that can be allocated from the
    #: subnet pool. *Type: int*
    maximum_prefix_length = resource.Body('max_prefixlen', type=int)
    #: The minimum prefix length that can be allocated from the
    #: subnet pool. *Type: int*
    minimum_prefix_length = resource.Body('min_prefixlen', type=int)
    #: The subnet pool name.
    name = resource.Body('name')
    #: The ID of the project that owns the subnet pool.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: A list of subnet prefixes that are assigned to the subnet pool.
    #: The adjacent prefixes are merged and treated as a single prefix.
    #: *Type: list*
    prefixes = resource.Body('prefixes', type=list)
    #: Revision number of the subnet pool. *Type: int*
    revision_number = resource.Body('revision_number', type=int)
    #: Timestamp when the subnet pool was last updated.
    updated_at = resource.Body('updated_at')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import port_binding
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'host': 'host1',
    'profile': {},
    'vif_details': {'bridge_name': 'br-int'},
    'vif_type': 'ovs',
    'vnic_type': 'normal',
}


class TestPortBinding(base.TestCase):
    def test_basic(self):
        sot = port_binding.PortBinding()
        self.assertEqual('binding', sot.resource_key)
        self.assertEqual('bindings', sot.resources_key)
        self.assertEqual('/ports/%(port_id)s/bindings', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = port_binding.PortBinding(**EXAMPLE)
        self.assertEqual(EXAMPLE['host'], sot.host)
        self.assertEqual(EXAMPLE['profile'], sot.profile)
        self.assertEqual(EXAMPLE['vif_details'], sot.vif_details)
        self.assertEqual(EXAMPLE['vif_type'], sot.vif_type)
        self.assertCountEqual(EXAMPLE['vnic_type'], sot.vnic_type)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.tests import fakes
from openstack.tests.unit import base


class TestMicroversions(base.TestCase):
    def setUp(self):
        super().setUp()
        self.use_compute_discovery()

    def test_get_bad_inferred_max_microversion(self):
        self.cloud.config.config['compute_api_version'] = '2.61'

        self.assertRaises(
            exceptions.ConfigException,
            self.cloud.get_server,
            'doesNotExist',
        )

        self.assert_calls()

    def test_get_bad_default_max_microversion(self):
        self.cloud.config.config['compute_default_microversion'] = '2.61'

        self.assertRaises(
            exceptions.ConfigException,
            self.cloud.get_server,
            'doesNotExist',
        )

        self.assert_calls()

    def test_get_bad_inferred_min_microversion(self):
        self.cloud.config.config['compute_api_version'] = '2.7'

        self.assertRaises(
            exceptions.ConfigException,
            self.cloud.get_server,
            'doesNotExist',
        )

        self.assert_calls()

    def test_get_bad_default_min_microversion(self):
        self.cloud.config.config['compute_default_microversion'] = '2.7'

        self.assertRaises(
            exceptions.ConfigException,
            self.cloud.get_server,
            'doesNotExist',
        )

        self.assert_calls()

    def test_inferred_default_microversion(self):
        self.cloud.config.config['compute_api_version'] = '2.42'

        server = fakes.make_fake_server('123', 'mickey')

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', 'mickey']
                    ),
                    request_headers={'OpenStack-API-Version': 'compute 2.42'},
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', 'detail'],
                        qs_elements=['name=mickey'],
                    ),
                    request_headers={'OpenStack-API-Version': 'compute 2.42'},
                    json={'servers': [server]},
                ),
            ]
        )

        r = self.cloud.get_server('mickey', bare=True)
        self.assertIsNotNone(r)
        self.assertEqual(server['name'], r['name'])

        self.assert_calls()

    def test_default_microversion(self):
        self.cloud.config.config['compute_default_microversion'] = '2.42'

        server = fakes.make_fake_server('123', 'mickey')

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', 'mickey']
                    ),
                    request_headers={'OpenStack-API-Version': 'compute 2.42'},
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', 'detail'],
                        qs_elements=['name=mickey'],
                    ),
                    request_headers={'OpenStack-API-Version': 'compute 2.42'},
                    json={'servers': [server]},
                ),
            ]
        )

        r = self.cloud.get_server('mickey', bare=True)
        self.assertIsNotNone(r)
        self.assertEqual(server['name'], r['name'])

        self.assert_calls()

    def test_conflicting_implied_and_direct(self):
        self.cloud.config.config['compute_default_microversion'] = '2.7'
        self.cloud.config.config['compute_api_version'] = '2.13'

        self.assertRaises(exceptions.ConfigException, self.cloud.get_server)

        # We should fail before we even authenticate
        self.assertEqual(0, len(self.adapter.request_history))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base


class TestAccount(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('object-store')

    def tearDown(self):
        account = self.operator_cloud.object_store.get_account_metadata()
        self.operator_cloud.object_store.delete_account_metadata(
            account.metadata.keys()
        )
        super().tearDown()

    def test_system_metadata(self):
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertGreaterEqual(account.account_bytes_used, 0)
        self.assertGreaterEqual(account.account_container_count, 0)
        self.assertGreaterEqual(account.account_object_count, 0)

    def test_custom_metadata(self):
        # get custom metadata
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertFalse(account.metadata)

        # set no custom metadata
        self.operator_cloud.object_store.set_account_metadata()
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertFalse(account.metadata)

        # set empty custom metadata
        self.operator_cloud.object_store.set_account_metadata(k0='')
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertFalse(account.metadata)

        # set custom metadata
        self.operator_cloud.object_store.set_account_metadata(k1='v1')
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertTrue(account.metadata)
        self.assertEqual(1, len(account.metadata))
        self.assertIn('k1', account.metadata)
        self.assertEqual('v1', account.metadata['k1'])

        # set more custom metadata
        self.operator_cloud.object_store.set_account_metadata(k2='v2')
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertTrue(account.metadata)
        self.assertEqual(2, len(account.metadata))
        self.assertIn('k1', account.metadata)
        self.assertEqual('v1', account.metadata['k1'])
        self.assertIn('k2', account.metadata)
        self.assertEqual('v2', account.metadata['k2'])

        # update custom metadata
        self.operator_cloud.object_store.set_account_metadata(k1='v1.1')
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertTrue(account.metadata)
        self.assertEqual(2, len(account.metadata))
        self.assertIn('k1', account.metadata)
        self.assertEqual('v1.1', account.metadata['k1'])
        self.assertIn('k2', account.metadata)
        self.assertEqual('v2', account.metadata['k2'])

        # unset custom metadata
        self.operator_cloud.object_store.delete_account_metadata(['k1'])
        account = self.operator_cloud.object_store.get_account_metadata()
        self.assertTrue(account.metadata)
        self.assertEqual(1, len(account.metadata))
        self.assertIn('k2', account.metadata)
        self.assertEqual('v2', account.metadata['k2'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
from uuid import uuid4

import testtools

from openstack.cloud import _utils
from openstack import exceptions
from openstack.tests.unit import base

RANGE_DATA = [
    dict(id=1, key1=1, key2=5),
    dict(id=2, key1=1, key2=20),
    dict(id=3, key1=2, key2=10),
    dict(id=4, key1=2, key2=30),
    dict(id=5, key1=3, key2=40),
    dict(id=6, key1=3, key2=40),
]


class TestUtils(base.TestCase):
    def test__filter_list_name_or_id(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'donald', None)
        self.assertEqual([el1], ret)

    def test__filter_list_name_or_id_special(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto[2017-01-10]')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'pluto[2017-01-10]', None)
        self.assertEqual([el2], ret)

    def test__filter_list_name_or_id_partial_bad(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto[2017-01-10]')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'pluto[2017-01]', None)
        self.assertEqual([], ret)

    def test__filter_list_name_or_id_partial_glob(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto[2017-01-10]')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'pluto*', None)
        self.assertEqual([el2], ret)

    def test__filter_list_name_or_id_non_glob_glob(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto[2017-01-10]')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'pluto', None)
        self.assertEqual([], ret)

    def test__filter_list_name_or_id_glob(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto')
        el3 = dict(id=200, name='pluto-2')
        data = [el1, el2, el3]
        ret = _utils._filter_list(data, 'pluto*', None)
        self.assertEqual([el2, el3], ret)

    def test__filter_list_name_or_id_glob_not_found(self):
        el1 = dict(id=100, name='donald')
        el2 = dict(id=200, name='pluto')
        el3 = dict(id=200, name='pluto-2')
        data = [el1, el2, el3]
        ret = _utils._filter_list(data, 'q*', None)
        self.assertEqual([], ret)

    def test__filter_list_unicode(self):
        el1 = dict(
            id=100,
            name='中文',
            last='duck',
            other=dict(category='duck', financial=dict(status='poor')),
        )
        el2 = dict(
            id=200,
            name='中文',
            last='trump',
            other=dict(category='human', financial=dict(status='rich')),
        )
        el3 = dict(
            id=300,
            name='donald',
            last='ronald mac',
            other=dict(category='clown', financial=dict(status='rich')),
        )
        data = [el1, el2, el3]
        ret = _utils._filter_list(
            data, '中文', {'other': {'financial': {'status': 'rich'}}}
        )
        self.assertEqual([el2], ret)

    def test__filter_list_filter(self):
        el1 = dict(id=100, name='donald', other='duck')
        el2 = dict(id=200, name='donald', other='trump')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'donald', {'other': 'duck'})
        self.assertEqual([el1], ret)

    def test__filter_list_filter_jmespath(self):
        el1 = dict(id=100, name='donald', other='duck')
        el2 = dict(id=200, name='donald', other='trump')
        data = [el1, el2]
        ret = _utils._filter_list(data, 'donald', "[?other == `duck`]")
        self.assertEqual([el1], ret)

    def test__filter_list_dict1(self):
        el1 = dict(
            id=100, name='donald', last='duck', other=dict(category='duck')
        )
        el2 = dict(
            id=200, name='donald', last='trump', other=dict(category='human')
        )
        el3 = dict(
            id=300,
            name='donald',
            last='ronald mac',
            other=dict(category='clown'),
        )
        data = [el1, el2, el3]
        ret = _utils._filter_list(
            data, 'donald', {'other': {'category': 'clown'}}
        )
        self.assertEqual([el3], ret)

    def test__filter_list_dict2(self):
        el1 = dict(
            id=100,
            name='donald',
            last='duck',
            other=dict(category='duck', financial=dict(status='poor')),
        )
        el2 = dict(
            id=200,
            name='donald',
            last='trump',
            other=dict(category='human', financial=dict(status='rich')),
        )
        el3 = dict(
            id=300,
            name='donald',
            last='ronald mac',
            other=dict(category='clown', financial=dict(status='rich')),
        )
        data = [el1, el2, el3]
        ret = _utils._filter_list(
            data, 'donald', {'other': {'financial': {'status': 'rich'}}}
        )
        self.assertEqual([el2, el3], ret)

    def test_safe_dict_min_ints(self):
        """Test integer comparison"""
        data = [{'f1': 3}, {'f1': 2}, {'f1': 1}]
        retval = _utils.safe_dict_min('f1', data)
        self.assertEqual(1, retval)

    def test_safe_dict_min_strs(self):
        """Test integer as strings comparison"""
        data = [{'f1': '3'}, {'f1': '2'}, {'f1': '1'}]
        retval = _utils.safe_dict_min('f1', data)
        self.assertEqual(1, retval)

    def test_safe_dict_min_None(self):
        """Test None values"""
        data = [{'f1': 3}, {'f1': None}, {'f1': 1}]
        retval = _utils.safe_dict_min('f1', data)
        self.assertEqual(1, retval)

    def test_safe_dict_min_key_missing(self):
        """Test missing key for an entry still works"""
        data = [{'f1': 3}, {'x': 2}, {'f1': 1}]
        retval = _utils.safe_dict_min('f1', data)
        self.assertEqual(1, retval)

    def test_safe_dict_min_key_not_found(self):
        """Test key not found in any elements returns None"""
        data = [{'f1': 3}, {'f1': 2}, {'f1': 1}]
        retval = _utils.safe_dict_min('doesnotexist', data)
        self.assertIsNone(retval)

    def test_safe_dict_min_not_int(self):
        """Test non-integer key value raises OSCE"""
        data = [{'f1': 3}, {'f1': "aaa"}, {'f1': 1}]
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Search for minimum value failed. "
            "Value for f1 is not an integer: aaa",
        ):
            _utils.safe_dict_min('f1', data)

    def test_safe_dict_max_ints(self):
        """Test integer comparison"""
        data = [{'f1': 3}, {'f1': 2}, {'f1': 1}]
        retval = _utils.safe_dict_max('f1', data)
        self.assertEqual(3, retval)

    def test_safe_dict_max_strs(self):
        """Test integer as strings comparison"""
        data = [{'f1': '3'}, {'f1': '2'}, {'f1': '1'}]
        retval = _utils.safe_dict_max('f1', data)
        self.assertEqual(3, retval)

    def test_safe_dict_max_None(self):
        """Test None values"""
        data = [{'f1': 3}, {'f1': None}, {'f1': 1}]
        retval = _utils.safe_dict_max('f1', data)
        self.assertEqual(3, retval)

    def test_safe_dict_max_key_missing(self):
        """Test missing key for an entry still works"""
        data = [{'f1': 3}, {'x': 2}, {'f1': 1}]
        retval = _utils.safe_dict_max('f1', data)
        self.assertEqual(3, retval)

    def test_safe_dict_max_key_not_found(self):
        """Test key not found in any elements returns None"""
        data = [{'f1': 3}, {'f1': 2}, {'f1': 1}]
        retval = _utils.safe_dict_max('doesnotexist', data)
        self.assertIsNone(retval)

    def test_safe_dict_max_not_int(self):
        """Test non-integer key value raises OSCE"""
        data = [{'f1': 3}, {'f1': "aaa"}, {'f1': 1}]
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Search for maximum value failed. "
            "Value for f1 is not an integer: aaa",
        ):
            _utils.safe_dict_max('f1', data)

    def test_parse_range_None(self):
        self.assertIsNone(_utils.parse_range(None))

    def test_parse_range_invalid(self):
        self.assertIsNone(_utils.parse_range("<invalid"))

    def test_parse_range_int_only(self):
        retval = _utils.parse_range("1024")
        self.assertIsInstance(retval, tuple)
        self.assertIsNone(retval[0])
        self.assertEqual(1024, retval[1])

    def test_parse_range_lt(self):
        retval = _utils.parse_range("<1024")
        self.assertIsInstance(retval, tuple)
        self.assertEqual("<", retval[0])
        self.assertEqual(1024, retval[1])

    def test_parse_range_gt(self):
        retval = _utils.parse_range(">1024")
        self.assertIsInstance(retval, tuple)
        self.assertEqual(">", retval[0])
        self.assertEqual(1024, retval[1])

    def test_parse_range_le(self):
        retval = _utils.parse_range("<=1024")
        self.assertIsInstance(retval, tuple)
        self.assertEqual("<=", retval[0])
        self.assertEqual(1024, retval[1])

    def test_parse_range_ge(self):
        retval = _utils.parse_range(">=1024")
        self.assertIsInstance(retval, tuple)
        self.assertEqual(">=", retval[0])
        self.assertEqual(1024, retval[1])

    def test_range_filter_min(self):
        retval = _utils.range_filter(RANGE_DATA, "key1", "min")
        self.assertIsInstance(retval, list)
        self.assertEqual(2, len(retval))
        self.assertEqual(RANGE_DATA[:2], retval)

    def test_range_filter_max(self):
        retval = _utils.range_filter(RANGE_DATA, "key1", "max")
        self.assertIsInstance(retval, list)
        self.assertEqual(2, len(retval))
        self.assertEqual(RANGE_DATA[-2:], retval)

    def test_range_filter_range(self):
        retval = _utils.range_filter(RANGE_DATA, "key1", "<3")
        self.assertIsInstance(retval, list)
        self.assertEqual(4, len(retval))
        self.assertEqual(RANGE_DATA[:4], retval)

    def test_range_filter_exact(self):
        retval = _utils.range_filter(RANGE_DATA, "key1", "2")
        self.assertIsInstance(retval, list)
        self.assertEqual(2, len(retval))
        self.assertEqual(RANGE_DATA[2:4], retval)

    def test_range_filter_invalid_int(self):
        with testtools.ExpectedException(
            exceptions.SDKException, "Invalid range value: <1A0"
        ):
            _utils.range_filter(RANGE_DATA, "key1", "<1A0")

    def test_range_filter_invalid_op(self):
        with testtools.ExpectedException(
            exceptions.SDKException, "Invalid range value: <>100"
        ):
            _utils.range_filter(RANGE_DATA, "key1", "<>100")

    def test_get_entity_pass_object(self):
        obj = mock.Mock(id=uuid4().hex)
        self.cloud.use_direct_get = True
        self.assertEqual(obj, _utils._get_entity(self.cloud, '', obj, {}))

    def test_get_entity_pass_dict(self):
        d = dict(id=uuid4().hex)
        self.cloud.use_direct_get = True
        self.assertEqual(d, _utils._get_entity(self.cloud, '', d, {}))

    def test_get_entity_no_use_direct_get(self):
        # test we are defaulting to the search_<resource> methods
        # if the use_direct_get flag is set to False(default).
        uuid = uuid4().hex
        resource = 'network'
        func = f'search_{resource}s'
        filters = {}
        with mock.patch.object(self.cloud, func) as search:
            _utils._get_entity(self.cloud, resource, uuid, filters)
            search.assert_called_once_with(uuid, filters)

    def test_get_entity_no_uuid_like(self):
        # test we are defaulting to the search_<resource> methods
        # if the name_or_id param is a name(string) but not a uuid.
        self.cloud.use_direct_get = True
        name = 'name_no_uuid'
        resource = 'network'
        func = f'search_{resource}s'
        filters = {}
        with mock.patch.object(self.cloud, func) as search:
            _utils._get_entity(self.cloud, resource, name, filters)
            search.assert_called_once_with(name, filters)

    def test_get_entity_pass_uuid(self):
        uuid = uuid4().hex
        self.cloud.use_direct_get = True
        resources = [
            'flavor',
            'image',
            'volume',
            'network',
            'subnet',
            'port',
            'floating_ip',
            'security_group',
        ]
        for r in resources:
            f = f'get_{r}_by_id'
            with mock.patch.object(self.cloud, f) as get:
                _utils._get_entity(self.cloud, r, uuid, {})
                get.assert_called_once_with(uuid)

    def test_get_entity_pass_search_methods(self):
        self.cloud.use_direct_get = True
        resources = [
            'flavor',
            'image',
            'volume',
            'network',
            'subnet',
            'port',
            'floating_ip',
            'security_group',
        ]
        filters = {}
        name = 'name_no_uuid'
        for r in resources:
            f = f'search_{r}s'
            with mock.patch.object(self.cloud, f) as search:
                _utils._get_entity(self.cloud, r, name, {})
                search.assert_called_once_with(name, filters)

    def test_get_entity_get_and_search(self):
        resources = [
            'flavor',
            'image',
            'volume',
            'network',
            'subnet',
            'port',
            'floating_ip',
            'security_group',
        ]
        for r in resources:
            self.assertTrue(hasattr(self.cloud, f'get_{r}_by_id'))
            self.assertTrue(hasattr(self.cloud, f'search_{r}s'))
